import hashlib
import logging
import os
from typing import Any, Optional

import requests

from mem0x.bootstrap.helpers.json_serializable import register_deserializable
from mem0x.bootstrap.loaders.base_loader import BaseLoader
from mem0x.bootstrap.utils.misc import clean_string

logger = logging.getLogger(__name__)


class NotionDocument:
    """
    A simple Document class to hold the text and additional information of a page.
    """

    def __init__(self, text: str, extra_info: dict[str, Any]):
        self.text = text
        self.extra_info = extra_info


class NotionPageLoader:
    """
    Notion Page Loader.
    Reads a set of Notion pages.
    """

    BLOCK_CHILD_URL_TMPL = "https://api.notion.com/v1/blocks/{block_id}/children"

    def __init__(self, integration_token: Optional[str] = None) -> None:
        """Initialize with Notion integration token."""
        if integration_token is None:
            integration_token = os.getenv("NOTION_INTEGRATION_TOKEN")
            if integration_token is None:
                raise ValueError(
                    "Must specify `integration_token` or set environment " "variable `NOTION_INTEGRATION_TOKEN`."
                )
        self.token = integration_token
        self.headers = {
            "Authorization": "Bearer " + self.token,
            "Content-Type": "application/json",
            "Notion-Version": "2022-06-28",
        }

    def _read_block(self, block_id: str, num_tabs: int = 0) -> str:
        """Read a block from Notion."""
        done = False
        result_lines_arr = []
        cur_block_id = block_id
        while not done:
            block_url = self.BLOCK_CHILD_URL_TMPL.format(block_id=cur_block_id)
            res = requests.get(block_url, headers=self.headers)
            data = res.json()

            for result in data["results"]:
                result_type = result["type"]
                result_obj = result[result_type]

                cur_result_text_arr = []
                if "rich_text" in result_obj:
                    for rich_text in result_obj["rich_text"]:
                        if "text" in rich_text:
                            text = rich_text["text"]["content"]
                            prefix = "\t" * num_tabs
                            cur_result_text_arr.append(prefix + text)

                result_block_id = result["id"]
                has_children = result["has_children"]
                if has_children:
                    children_text = self._read_block(result_block_id, num_tabs=num_tabs + 1)
                    cur_result_text_arr.append(children_text)

                cur_result_text = "\n".join(cur_result_text_arr)
                result_lines_arr.append(cur_result_text)

            if data["next_cursor"] is None:
                done = True
            else:
                cur_block_id = data["next_cursor"]

        result_lines = "\n".join(result_lines_arr)
        return result_lines

    def load_data(self, page_ids: list[str]) -> list[NotionDocument]:
        """Load data from the given list of page IDs."""
        docs = []
        for page_id in page_ids:
            page_text = self._read_block(page_id)
            docs.append(NotionDocument(text=page_text, extra_info={"page_id": page_id}))
        return docs


@register_deserializable
class NotionLoader(BaseLoader):
    def load_data(self, source):
        """Load data from a Notion URL."""

        id = source[-32:]
        formatted_id = f"{id[:8]}-{id[8:12]}-{id[12:16]}-{id[16:20]}-{id[20:]}"
        logger.debug(f"Extracted notion page id as: {formatted_id}")

        integration_token = os.getenv("NOTION_INTEGRATION_TOKEN")
        reader = NotionPageLoader(integration_token=integration_token)
        documents = reader.load_data(page_ids=[formatted_id])

        raw_text = documents[0].text

        text = clean_string(raw_text)
        doc_id = hashlib.sha256((text + source).encode()).hexdigest()
        return {
            "doc_id": doc_id,
            "data": [
                {
                    "content": text,
                    "meta_data": {"url": f"notion-{formatted_id}"},
                }
            ],
        }



import base64
import hashlib
import os
from pathlib import Path

from openai import OpenAI

from mem0x.bootstrap.helpers.json_serializable import register_deserializable
from mem0x.bootstrap.loaders.base_loader import BaseLoader

DESCRIBE_IMAGE_PROMPT = "Describe the image:"


@register_deserializable
class ImageLoader(BaseLoader):
    def __init__(self, max_tokens: int = 500, api_key: str = None, prompt: str = None):
        super().__init__()
        self.custom_prompt = prompt or DESCRIBE_IMAGE_PROMPT
        self.max_tokens = max_tokens
        self.api_key = api_key or os.environ["OPENAI_API_KEY"]
        self.client = OpenAI(api_key=self.api_key)

    @staticmethod
    def _encode_image(image_path: str):
        with open(image_path, "rb") as image_file:
            return base64.b64encode(image_file.read()).decode("utf-8")

    def _create_completion_request(self, content: str):
        return self.client.chat.completions.create(
            model="gpt-4o", messages=[{"role": "user", "content": content}], max_tokens=self.max_tokens
        )

    def _process_url(self, url: str):
        if url.startswith("http"):
            return [{"type": "text", "text": self.custom_prompt}, {"type": "image_url", "image_url": {"url": url}}]
        elif Path(url).is_file():
            extension = Path(url).suffix.lstrip(".")
            encoded_image = self._encode_image(url)
            image_data = f"data:image/{extension};base64,{encoded_image}"
            return [{"type": "text", "text": self.custom_prompt}, {"type": "image", "image_url": {"url": image_data}}]
        else:
            raise ValueError(f"Invalid URL or file path: {url}")

    def load_data(self, url: str):
        content = self._process_url(url)
        response = self._create_completion_request(content)
        content = response.choices[0].message.content

        doc_id = hashlib.sha256((content + url).encode()).hexdigest()
        return {"doc_id": doc_id, "data": [{"content": content, "meta_data": {"url": url, "type": "image"}}]}



import hashlib
import json
import logging

try:
    from youtube_transcript_api import YouTubeTranscriptApi
except ImportError:
    raise ImportError("YouTube video requires extra dependencies. Install with `pip install youtube-transcript-api`")
try:
    from langchain_community.document_loaders import YoutubeLoader
    from langchain_community.document_loaders.youtube import _parse_video_id
except ImportError:
    raise ImportError("YouTube video requires extra dependencies. Install with `pip install pytube==15.0.0`") from None
from mem0x.bootstrap.helpers.json_serializable import register_deserializable
from mem0x.bootstrap.loaders.base_loader import BaseLoader
from mem0x.bootstrap.utils.misc import clean_string


@register_deserializable
class YoutubeVideoLoader(BaseLoader):
    def load_data(self, url):
        """Load data from a Youtube video."""
        video_id = _parse_video_id(url)

        languages = ["en"]
        try:
            # Fetching transcript data
            languages = [transcript.language_code for transcript in YouTubeTranscriptApi.list_transcripts(video_id)]
            transcript = YouTubeTranscriptApi.get_transcript(video_id, languages=languages)
            # convert transcript to json to avoid unicode symboles
            transcript = json.dumps(transcript, ensure_ascii=True)
        except Exception:
            logging.exception(f"Failed to fetch transcript for video {url}")
            transcript = "Unavailable"

        loader = YoutubeLoader.from_youtube_url(url, add_video_info=True, language=languages)
        doc = loader.load()
        output = []
        if not len(doc):
            raise ValueError(f"No data found for url: {url}")
        content = doc[0].page_content
        content = clean_string(content)
        metadata = doc[0].metadata
        metadata["url"] = url
        metadata["transcript"] = transcript

        output.append(
            {
                "content": content,
                "meta_data": metadata,
            }
        )
        doc_id = hashlib.sha256((content + url).encode()).hexdigest()
        return {
            "doc_id": doc_id,
            "data": output,
        }



from datetime import datetime, timedelta

from fastapi import FastAPI

from mem0x.app.instance import memoryInstance
from mem0x.app.type import (BootstrapRequest, LongTermAddRequest,
                            LongTermRemoveAllRequest, LongTermRemoveRequest,
                            LongTermSearchRequest, LongTermUpdateRequest,
                            Response, ShortTermAddRequest,
                            ShortTermDeleteRequest, ShortTermGetRequest,
                            ShortTermGetRunIdsRequest)
from mem0x.bootstrap.config import AddConfig, ChunkerConfig
from mem0x.bootstrap.data_formatter import DataFormatter
from mem0x.bootstrap.models.data_type import DataType
from mem0x.configs.prompts import (FACT_RETRIEVAL_PROMPT,
                                   FACT_RETRIEVAL_PROMPT_KNOWLEDGE)

app = FastAPI()


@app.post("/longterm_search")
async def longterm_search(search_request: LongTermSearchRequest):
    try:
        filters = {"table_name": search_request.table_name}
        related_memories = memoryInstance.m.search(
            search_request.query,
            search_request.user_id,
            search_request.agent_id,
            search_request.run_id,
            search_request.limit,
            filters,
        )
        return Response(message="SUCCESS", status=True, data=related_memories)
    except Exception as e:
        return Response(message=f"{e}", status=False, data=None)


@app.post("/longterm_add")
async def longterm_add(add_request: LongTermAddRequest):
    try:
        print(add_request.messages)
        metadata = {"table_name": add_request.table_name}
        filters = {"table_name": add_request.table_name}
        result = memoryInstance.m.add(
            add_request.messages,
            add_request.user_id,
            add_request.agent_id,
            add_request.run_id,
            metadata,
            filters,
            FACT_RETRIEVAL_PROMPT,
        )
        return Response(message="SUCCESS", status=True, data=result)
    except Exception as e:
        return Response(message=f"{e}", status=False, data=None)


@app.post("/longterm_remove")
async def longterm_remove(remove_request: LongTermRemoveRequest):
    try:
        memoryInstance.m.delete(remove_request.memory_id)
        return Response(message="SUCCESS", status=True, data=None)
    except Exception as e:
        return Response(message=f"{e}", status=False, data=None)


@app.post("/longterm_remove_all")
async def longterm_remove_all(remove_all_request: LongTermRemoveAllRequest):
    try:
        memoryInstance.m.delete_all(remove_all_request.user_id, remove_all_request.agent_id, remove_all_request.run_id)
        return Response(message="SUCCESS", status=True, data=None)
    except Exception as e:
        return Response(message=f"{e}", status=False, data=None)


@app.post("/longterm_update")
async def longterm_update(update_request: LongTermUpdateRequest):
    try:
        memoryInstance.m.update(update_request.memory_id, update_request.data)
        return Response(message="SUCCESS", status=True, data=None)
    except Exception as e:
        return Response(message=f"{e}", status=False, data=None)


@app.post("/bootstrap_add")
async def bootstrap_add(bootstrap_request: BootstrapRequest):
    try:
        loader, chunker = None, None
        config = AddConfig()
        config.chunker = ChunkerConfig(
            chunk_size=bootstrap_request.chunk_size,
            chunk_overlap=bootstrap_request.chunk_overlap,
            min_chunk_size=bootstrap_request.min_chunk_size,
        )
        data_type = DataType(bootstrap_request.data_type)
        data_formatter = DataFormatter(data_type, config, loader, chunker)

        src = bootstrap_request.src
        embeddings_data = data_formatter.chunker.create_chunks(data_formatter.loader, src)

        messages = []
        for document in embeddings_data["documents"]:
            messages.append({"role": "", "content": document})

        metadata = {"table_name": bootstrap_request.table_name}
        filters = {"table_name": bootstrap_request.table_name}
        result = memoryInstance.m.add(
            messages,
            bootstrap_request.user_id,
            bootstrap_request.agent_id,
            bootstrap_request.run_id,
            metadata,
            filters,
            FACT_RETRIEVAL_PROMPT_KNOWLEDGE,
        )
        return Response(message="SUCCESS", status=True, data=result)
    except Exception as e:
        return Response(message=f"{e}", status=False, data=None)


@app.post("/shortterm_add")
async def shortterm_add(add_request: ShortTermAddRequest):
    try:
        memoryInstance.sm.add(
            add_request.content, add_request.user_id, add_request.agent_id, add_request.run_id, add_request.table_name
        )
        return Response(message="SUCCESS", status=True, data=None)
    except Exception as e:
        return Response(message=f"{e}", status=False, data=None)


@app.post("/shortterm_get")
async def shortterm_get(get_request: ShortTermGetRequest):
    try:
        timestamp = get_request.timestamp
        if timestamp is None:
            comparison_time = datetime.now() + timedelta(hours=24)
            timestamp = comparison_time.isoformat()

        result = memoryInstance.sm.get(get_request.table_name, get_request.run_id, timestamp)
        return Response(message="SUCCESS", status=True, data=result)
    except Exception as e:
        return Response(message=f"{e}", status=False, data=None)


@app.post("/shortterm_get_runids")
async def shortterm_get_runids(get_runids_request: ShortTermGetRunIdsRequest):
    try:
        timestamp = get_runids_request.timestamp
        if timestamp is None:
            comparison_time = datetime.now() + timedelta(hours=24)
            timestamp = comparison_time.isoformat()

        result = memoryInstance.sm.get_runids(timestamp)
        return Response(message="SUCCESS", status=True, data=result)
    except Exception as e:
        return Response(message=f"{e}", status=False, data=None)


@app.post("/shortterm_delete")
async def shortterm_delete(delete_request: ShortTermDeleteRequest):
    try:
        memoryInstance.sm.delete(delete_request.id)
        return Response(message="SUCCESS", status=True, data=None)
    except Exception as e:
        return Response(message=f"{e}", status=False, data=None)



import builtins
import logging
from collections.abc import Callable
from importlib import import_module
from typing import Optional

from mem0x.bootstrap.config.base_config import BaseConfig
from mem0x.bootstrap.helpers.json_serializable import register_deserializable


@register_deserializable
class ChunkerConfig(BaseConfig):
    """
    Config for the chunker used in `add` method
    """

    def __init__(
        self,
        chunk_size: Optional[int] = 2000,
        chunk_overlap: Optional[int] = 0,
        length_function: Optional[Callable[[str], int]] = None,
        min_chunk_size: Optional[int] = 0,
    ):
        self.chunk_size = chunk_size
        self.chunk_overlap = chunk_overlap
        self.min_chunk_size = min_chunk_size
        if self.min_chunk_size >= self.chunk_size:
            raise ValueError(f"min_chunk_size {min_chunk_size} should be less than chunk_size {chunk_size}")
        if self.min_chunk_size < self.chunk_overlap:
            logging.warning(
                f"min_chunk_size {min_chunk_size} should be greater than chunk_overlap {chunk_overlap}, otherwise it is redundant."  # noqa:E501
            )

        if isinstance(length_function, str):
            self.length_function = self.load_func(length_function)
        else:
            self.length_function = length_function if length_function else len

    @staticmethod
    def load_func(dotpath: str):
        if "." not in dotpath:
            return getattr(builtins, dotpath)
        else:
            module_, func = dotpath.rsplit(".", maxsplit=1)
            m = import_module(module_)
            return getattr(m, func)


@register_deserializable
class LoaderConfig(BaseConfig):
    """
    Config for the loader used in `add` method
    """

    def __init__(self):
        pass


@register_deserializable
class AddConfig(BaseConfig):
    """
    Config for the `add` method.
    """

    def __init__(
        self,
        chunker: Optional[ChunkerConfig] = None,
        loader: Optional[LoaderConfig] = None,
    ):
        """
        Initializes a configuration class instance for the `add` method.

        :param chunker: Chunker config, defaults to None
        :type chunker: Optional[ChunkerConfig], optional
        :param loader: Loader config, defaults to None
        :type loader: Optional[LoaderConfig], optional
        """
        self.loader = loader
        self.chunker = chunker



import hashlib
import logging
import os

from mem0x.bootstrap.helpers.json_serializable import register_deserializable
from mem0x.bootstrap.loaders.base_loader import BaseLoader

logger = logging.getLogger(__name__)


@register_deserializable
class DiscordLoader(BaseLoader):
    """
    Load data from a Discord Channel ID.
    """

    def __init__(self):
        if not os.environ.get("DISCORD_TOKEN"):
            raise ValueError("DISCORD_TOKEN is not set")

        self.token = os.environ.get("DISCORD_TOKEN")

    @staticmethod
    def _format_message(message):
        return {
            "message_id": message.id,
            "content": message.content,
            "author": {
                "id": message.author.id,
                "name": message.author.name,
                "discriminator": message.author.discriminator,
            },
            "created_at": message.created_at.isoformat(),
            "attachments": [
                {
                    "id": attachment.id,
                    "filename": attachment.filename,
                    "size": attachment.size,
                    "url": attachment.url,
                    "proxy_url": attachment.proxy_url,
                    "height": attachment.height,
                    "width": attachment.width,
                }
                for attachment in message.attachments
            ],
            "embeds": [
                {
                    "title": embed.title,
                    "type": embed.type,
                    "description": embed.description,
                    "url": embed.url,
                    "timestamp": embed.timestamp.isoformat(),
                    "color": embed.color,
                    "footer": {
                        "text": embed.footer.text,
                        "icon_url": embed.footer.icon_url,
                        "proxy_icon_url": embed.footer.proxy_icon_url,
                    },
                    "image": {
                        "url": embed.image.url,
                        "proxy_url": embed.image.proxy_url,
                        "height": embed.image.height,
                        "width": embed.image.width,
                    },
                    "thumbnail": {
                        "url": embed.thumbnail.url,
                        "proxy_url": embed.thumbnail.proxy_url,
                        "height": embed.thumbnail.height,
                        "width": embed.thumbnail.width,
                    },
                    "video": {
                        "url": embed.video.url,
                        "height": embed.video.height,
                        "width": embed.video.width,
                    },
                    "provider": {
                        "name": embed.provider.name,
                        "url": embed.provider.url,
                    },
                    "author": {
                        "name": embed.author.name,
                        "url": embed.author.url,
                        "icon_url": embed.author.icon_url,
                        "proxy_icon_url": embed.author.proxy_icon_url,
                    },
                    "fields": [
                        {
                            "name": field.name,
                            "value": field.value,
                            "inline": field.inline,
                        }
                        for field in embed.fields
                    ],
                }
                for embed in message.embeds
            ],
        }

    def load_data(self, channel_id: str):
        """Load data from a Discord Channel ID."""
        import discord

        messages = []

        class DiscordClient(discord.Client):
            async def on_ready(self) -> None:
                logger.info("Logged on as {0}!".format(self.user))
                try:
                    channel = self.get_channel(int(channel_id))
                    if not isinstance(channel, discord.TextChannel):
                        raise ValueError(
                            f"Channel {channel_id} is not a text channel. " "Only text channels are supported for now."
                        )
                    threads = {}

                    for thread in channel.threads:
                        threads[thread.id] = thread

                    async for message in channel.history(limit=None):
                        messages.append(DiscordLoader._format_message(message))
                        if message.id in threads:
                            async for thread_message in threads[message.id].history(limit=None):
                                messages.append(DiscordLoader._format_message(thread_message))

                except Exception as e:
                    logger.error(e)
                    await self.close()
                finally:
                    await self.close()

        intents = discord.Intents.default()
        intents.message_content = True
        client = DiscordClient(intents=intents)
        client.run(self.token)

        metadata = {
            "url": channel_id,
        }

        messages = str(messages)

        doc_id = hashlib.sha256((messages + channel_id).encode()).hexdigest()

        return {
            "doc_id": doc_id,
            "data": [
                {
                    "content": messages,
                    "meta_data": metadata,
                }
            ],
        }



import hashlib

from mem0x.bootstrap.helpers.json_serializable import register_deserializable
from mem0x.bootstrap.loaders.base_loader import BaseLoader


@register_deserializable
class LocalQnaPairLoader(BaseLoader):
    def load_data(self, content):
        """Load data from a local QnA pair."""
        question, answer = content
        content = f"Q: {question}\nA: {answer}"
        url = "local"
        metadata = {"url": url, "question": question}
        doc_id = hashlib.sha256((content + url).encode()).hexdigest()
        return {
            "doc_id": doc_id,
            "data": [
                {
                    "content": content,
                    "meta_data": metadata,
                }
            ],
        }



import json
import uuid
from datetime import datetime

import psycopg2

# from psycopg2 import sql
# from psycopg2.extras import RealDictCursor


class PostgreSQLManager:
    def __init__(self, dbname: str, user: str, password: str, host: str = "localhost", port: int = 5432):
        """Initialize the PostgreSQLManager with a connection to the database.

        Args:
            dbname (str): Name of the PostgreSQL database.
            user (str): Database user.
            password (str): Database password.
            host (str): Database host. Defaults to "localhost".
            port (int): Database port. Defaults to 5432.
        """
        self.connection = psycopg2.connect(dbname=dbname, user=user, password=password, host=host, port=port)
        self._create_shortterm_memories_table()

    def _create_shortterm_memories_table(self) -> None:
        """Create the shortterm memory table if does not exist"""
        with self.connection.cursor() as cursor:
            cursor.execute("""
               CREATE TABLE IF NOT EXISTS shortterm_memories (
               id UUID PRIMARY KEY,
               content TEXT,
               user_id TEXT,
               agent_id TEXT,
               run_id TEXT,
               table_name TEXT,
               created_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP
               );
            """)
            self.connection.commit()
            cursor.close()

    def add(
        self,
        content,
        user_id,
        agent_id,
        run_id,
        table_name,
        created_at=None,
    ):
        with self.connection.cursor() as cursor:
            cursor.execute(
                """
                INSERT INTO shortterm_memories (id, content, user_id, agent_id, run_id, table_name, created_at)
                VALUES (%s, %s, %s, %s, %s, %s, %s)
            """,
                (
                    str(uuid.uuid4()),
                    json.dumps(content),
                    user_id,
                    agent_id,
                    run_id,
                    table_name,
                    created_at or datetime.now(),
                ),
            )
            self.connection.commit()
            cursor.close()

    def get(self, table_name, run_id, timestamp):
        with self.connection.cursor() as cursor:
            cursor.execute(
                """
                    SELECT id, content, user_id, agent_id, run_id, created_at
                    FROM shortterm_memories
                    WHERE run_id = %s and table_name = %s and created_at <= %s
                    ORDER BY created_at ASC
                    """,
                (
                    run_id,
                    table_name,
                    timestamp,
                ),
            )
            self.connection.commit()
            rows = cursor.fetchall()
            cursor.close()
            return [
                {
                    "id": row[0],
                    "content": json.loads(row[1]),
                    "user_id": row[2],
                    "agent_id": row[3],
                    "run_id": row[4],
                    "created_at": row[5],
                }
                for row in rows
            ]

    def get_runids(self, timestamp):
        with self.connection.cursor() as cursor:
            cursor.execute(
                """
               SELECT DISTINCT run_id
               FROM shortterm_memories where created_at <= %s
               """,
                (timestamp,),
            )
            self.connection.commit()
            rows = cursor.fetchall()
            cursor.close()
            return [
                {
                    "run_id": row[0],
                }
                for row in rows
            ]

    def delete(self, id):
        with self.connection.cursor() as cursor:
            cursor.execute(
                """DELETE from shortterm_memories where id = %s""",
                (id,),
            )
            self.connection.commit()
            cursor.close()

    def reset(self) -> None:
        """Reset the shortterm_memories table by dropping and recreating it."""
        with self.connection.cursor() as cursor:
            cursor.execute("DROP TABLE IF EXISTS shortterm_memories;")
            self.connection.commit()
            cursor.close()
            self._create_shortterm_memories_table()

    def close(self) -> None:
        """Close the database connection."""
        self.connection.close()



import os

from pydantic import BaseModel, Field


class PostgresStorageConfig(BaseModel):
    dbname: str = Field(os.getenv("POSTGRES_DBNAME") or "mem0x", description="Mem0x database")
    user: str = Field(os.getenv("POSTGRES_USER") or "user", description="Postgres user")
    password: str = (Field(os.getenv("POSTGRES_PASSWORD") or "password", description="Postgres user password"),)
    host: str = (Field(os.getenv("POSTGRES_HOST") or "127.0.0.1", description="Postgres host"),)
    port: int = Field(5432, description="Postgres port")



import os
from typing import Any, Dict, Optional

from pydantic import BaseModel, Field

from mem0x.embeddings.configs import EmbedderConfig
from mem0x.graphs.configs import GraphStoreConfig
from mem0x.llms.configs import LlmConfig
from mem0x.memory.configs import PostgresStorageConfig
from mem0x.memory.setup import mem0x_dir
from mem0x.vector_stores.configs import VectorStoreConfig


class MemoryItem(BaseModel):
    id: str = Field(..., description="The unique identifier for the text data")
    memory: str = Field(
        ..., description="The memory deduced from the text data"
    )  # TODO After prompt changes from platform, update this
    hash: Optional[str] = Field(None, description="The hash of the memory")
    # The metadata value can be anything and not just string. Fix it
    metadata: Optional[Dict[str, Any]] = Field(None, description="Additional metadata for the text data")
    score: Optional[float] = Field(None, description="The score associated with the text data")
    created_at: Optional[str] = Field(None, description="The timestamp when the memory was created")
    updated_at: Optional[str] = Field(None, description="The timestamp when the memory was updated")


class MemoryConfig(BaseModel):
    vector_store: VectorStoreConfig = Field(
        description="Configuration for the vector store",
        default_factory=VectorStoreConfig,
    )
    llm: LlmConfig = Field(
        description="Configuration for the language model",
        default_factory=LlmConfig,
    )
    embedder: EmbedderConfig = Field(
        description="Configuration for the embedding model",
        default_factory=EmbedderConfig,
    )
    sqlite_db_path: str = Field(
        description="Path to the sqlite database",
        default=os.path.join(mem0x_dir, "history.db"),
    )
    postgres_storage_config: PostgresStorageConfig = Field(
        description="Postgres storage config",
        default=None,
    )
    graph_store: GraphStoreConfig = Field(
        description="Configuration for the graph",
        default_factory=GraphStoreConfig,
    )
    version: str = Field(
        description="The version of the API",
        default="v1.0",
    )
    enable_history: bool = Field(
        description="Enable history",
        default=False,
    )


class ShortTermMemoryConfig(BaseModel):
    sqlite_db_path: str = Field(
        description="Path to the sqlite database",
        default=os.path.join(mem0x_dir, "shortterm.db"),
    )
    postgres_storage_config: PostgresStorageConfig = Field(
        description="Postgres storage config",
        default=None,
    )


class AzureConfig(BaseModel):
    """
    Configuration settings for Azure.

    Args:
        api_key (str): The API key used for authenticating with the Azure service.
        azure_deployment (str): The name of the Azure deployment.
        azure_endpoint (str): The endpoint URL for the Azure service.
        api_version (str): The version of the Azure API being used.
        default_headers (Dict[str, str]): Headers to include in requests to the Azure API.
    """

    api_key: str = Field(
        description="The API key used for authenticating with the Azure service.",
        default=None,
    )
    azure_deployment: str = Field(description="The name of the Azure deployment.", default=None)
    azure_endpoint: str = Field(description="The endpoint URL for the Azure service.", default=None)
    api_version: str = Field(description="The version of the Azure API being used.", default=None)
    default_headers: Optional[Dict[str, str]] = Field(
        description="Headers to include in requests to the Azure API.", default=None
    )



import hashlib
import logging
import time
from xml.etree import ElementTree

import requests

from mem0x.bootstrap.helpers.json_serializable import register_deserializable
from mem0x.bootstrap.loaders.base_loader import BaseLoader
from mem0x.bootstrap.utils.misc import is_readable

logger = logging.getLogger(__name__)


@register_deserializable
class SubstackLoader(BaseLoader):
    """
    This loader is used to load data from Substack URLs.
    """

    def load_data(self, url: str):
        try:
            from bs4 import BeautifulSoup
            from bs4.builder import ParserRejectedMarkup
        except ImportError:
            raise ImportError(
                "Substack requires extra dependencies. Install with `pip install beautifulsoup4==4.12.3`"
            ) from None

        if not url.endswith("sitemap.xml"):
            url = url + "/sitemap.xml"

        output = []
        response = requests.get(url)

        try:
            response.raise_for_status()
        except requests.exceptions.HTTPError as e:
            raise ValueError(
                f"""
                Failed to load {url}: {e}. Please use the root substack URL. For example, https://example.substack.com
                """
            )

        try:
            ElementTree.fromstring(response.content)
        except ElementTree.ParseError:
            raise ValueError(
                f"""
                Failed to parse {url}. Please use the root substack URL. For example, https://example.substack.com
                """
            )

        soup = BeautifulSoup(response.text, "xml")
        links = [link.text for link in soup.find_all("loc") if link.parent.name == "url" and "/p/" in link.text]
        if len(links) == 0:
            links = [link.text for link in soup.find_all("loc") if "/p/" in link.text]

        doc_id = hashlib.sha256((" ".join(links) + url).encode()).hexdigest()

        def serialize_response(soup: BeautifulSoup):
            data = {}

            h1_els = soup.find_all("h1")
            if h1_els is not None and len(h1_els) > 0:
                data["title"] = h1_els[1].text

            description_el = soup.find("meta", {"name": "description"})
            if description_el is not None:
                data["description"] = description_el["content"]

            content_el = soup.find("div", {"class": "available-content"})
            if content_el is not None:
                data["content"] = content_el.text

            like_btn = soup.find("div", {"class": "like-button-container"})
            if like_btn is not None:
                no_of_likes_div = like_btn.find("div", {"class": "label"})
                if no_of_likes_div is not None:
                    data["no_of_likes"] = no_of_likes_div.text

            return data

        def load_link(link: str):
            try:
                substack_data = requests.get(link)
                substack_data.raise_for_status()

                soup = BeautifulSoup(substack_data.text, "html.parser")
                data = serialize_response(soup)
                data = str(data)
                if is_readable(data):
                    return data
                else:
                    logger.warning(f"Page is not readable (too many invalid characters): {link}")
            except ParserRejectedMarkup as e:
                logger.error(f"Failed to parse {link}: {e}")
            return None

        for link in links:
            data = load_link(link)
            if data:
                output.append({"content": data, "meta_data": {"url": link}})
            # TODO: allow users to configure this
            time.sleep(1.0)  # added to avoid rate limiting

        return {"doc_id": doc_id, "data": output}



import datetime
import itertools
import json
import logging
import os
import re
import string
from typing import Any

from schema import Optional, Or, Schema
from tqdm import tqdm

from mem0x.bootstrap.models.data_type import DataType

logger = logging.getLogger(__name__)


def parse_content(content, type):
    implemented = ["html.parser", "lxml", "lxml-xml", "xml", "html5lib"]
    if type not in implemented:
        raise ValueError(f"Parser type {type} not implemented. Please choose one of {implemented}")

    from bs4 import BeautifulSoup

    soup = BeautifulSoup(content, type)
    original_size = len(str(soup.get_text()))

    tags_to_exclude = [
        "nav",
        "aside",
        "form",
        "header",
        "noscript",
        "svg",
        "canvas",
        "footer",
        "script",
        "style",
    ]
    for tag in soup(tags_to_exclude):
        tag.decompose()

    ids_to_exclude = ["sidebar", "main-navigation", "menu-main-menu"]
    for id in ids_to_exclude:
        tags = soup.find_all(id=id)
        for tag in tags:
            tag.decompose()

    classes_to_exclude = [
        "elementor-location-header",
        "navbar-header",
        "nav",
        "header-sidebar-wrapper",
        "blog-sidebar-wrapper",
        "related-posts",
    ]
    for class_name in classes_to_exclude:
        tags = soup.find_all(class_=class_name)
        for tag in tags:
            tag.decompose()

    content = soup.get_text()
    content = clean_string(content)

    cleaned_size = len(content)
    if original_size != 0:
        logger.info(
            f"Cleaned page size: {cleaned_size} characters, down from {original_size} (shrunk: {original_size-cleaned_size} chars, {round((1-(cleaned_size/original_size)) * 100, 2)}%)"  # noqa:E501
        )

    return content


def clean_string(text):
    """
    This function takes in a string and performs a series of text cleaning operations.

    Args:
        text (str): The text to be cleaned. This is expected to be a string.

    Returns:
        cleaned_text (str): The cleaned text after all the cleaning operations
        have been performed.
    """
    # Stripping and reducing multiple spaces to single:
    cleaned_text = re.sub(r"\s+", " ", text.strip())

    # Removing backslashes:
    cleaned_text = cleaned_text.replace("\\", "")

    # Replacing hash characters:
    cleaned_text = cleaned_text.replace("#", " ")

    # Eliminating consecutive non-alphanumeric characters:
    # This regex identifies consecutive non-alphanumeric characters (i.e., not
    # a word character [a-zA-Z0-9_] and not a whitespace) in the string
    # and replaces each group of such characters with a single occurrence of
    # that character.
    # For example, "!!! hello !!!" would become "! hello !".
    cleaned_text = re.sub(r"([^\w\s])\1*", r"\1", cleaned_text)

    return cleaned_text


def is_readable(s):
    """
    Heuristic to determine if a string is "readable" (mostly contains printable characters and forms meaningful words)

    :param s: string
    :return: True if the string is more than 95% printable.
    """
    len_s = len(s)
    if len_s == 0:
        return False
    printable_chars = set(string.printable)
    printable_ratio = sum(c in printable_chars for c in s) / len_s
    return printable_ratio > 0.95  # 95% of characters are printable


def use_pysqlite3():
    """
    Swap std-lib sqlite3 with pysqlite3.
    """
    import platform
    import sqlite3

    if platform.system() == "Linux" and sqlite3.sqlite_version_info < (3, 35, 0):
        try:
            # According to the Chroma team, this patch only works on Linux
            import datetime
            import subprocess
            import sys

            subprocess.check_call(
                [sys.executable, "-m", "pip", "install", "pysqlite3-binary", "--quiet", "--disable-pip-version-check"]
            )

            __import__("pysqlite3")
            sys.modules["sqlite3"] = sys.modules.pop("pysqlite3")

            # Let the user know what happened.
            current_time = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S,%f")[:-3]
            print(
                f"{current_time} [bootstrap] [INFO]",
                "Swapped std-lib sqlite3 with pysqlite3 for ChromaDb compatibility.",
                f"Your original version was {sqlite3.sqlite_version}.",
            )
        except Exception as e:
            # Escape all exceptions
            current_time = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S,%f")[:-3]
            print(
                f"{current_time} [bootstrap] [ERROR]",
                "Failed to swap std-lib sqlite3 with pysqlite3 for ChromaDb compatibility.",
                "Error:",
                e,
            )


def format_source(source: str, limit: int = 20) -> str:
    """
    Format a string to only take the first x and last x letters.
    This makes it easier to display a URL, keeping familiarity while ensuring a consistent length.
    If the string is too short, it is not sliced.
    """
    if len(source) > 2 * limit:
        return source[:limit] + "..." + source[-limit:]
    return source


def detect_datatype(source: Any) -> DataType:
    """
    Automatically detect the datatype of the given source.

    :param source: the source to base the detection on
    :return: data_type string
    """
    from urllib.parse import urlparse

    import requests
    import yaml

    def is_openapi_yaml(yaml_content):
        # currently the following two fields are required in openapi spec yaml config
        return "openapi" in yaml_content and "info" in yaml_content

    def is_google_drive_folder(url):
        # checks if url is a Google Drive folder url against a regex
        regex = r"^drive\.google\.com\/drive\/(?:u\/\d+\/)folders\/([a-zA-Z0-9_-]+)$"
        return re.match(regex, url)

    try:
        if not isinstance(source, str):
            raise ValueError("Source is not a string and thus cannot be a URL.")
        url = urlparse(source)
        # Check if both scheme and netloc are present. Local file system URIs are acceptable too.
        if not all([url.scheme, url.netloc]) and url.scheme != "file":
            raise ValueError("Not a valid URL.")
    except ValueError:
        url = False

    formatted_source = format_source(str(source), 30)

    if url:
        YOUTUBE_ALLOWED_NETLOCKS = {
            "www.youtube.com",
            "m.youtube.com",
            "youtu.be",
            "youtube.com",
            "vid.plus",
            "www.youtube-nocookie.com",
        }

        if url.netloc in YOUTUBE_ALLOWED_NETLOCKS:
            logger.debug(f"Source of `{formatted_source}` detected as `youtube_video`.")
            return DataType.YOUTUBE_VIDEO

        if url.netloc in {"notion.so", "notion.site"}:
            logger.debug(f"Source of `{formatted_source}` detected as `notion`.")
            return DataType.NOTION

        if url.path.endswith(".pdf"):
            logger.debug(f"Source of `{formatted_source}` detected as `pdf_file`.")
            return DataType.PDF_FILE

        if url.path.endswith(".xml"):
            logger.debug(f"Source of `{formatted_source}` detected as `sitemap`.")
            return DataType.SITEMAP

        if url.path.endswith(".csv"):
            logger.debug(f"Source of `{formatted_source}` detected as `csv`.")
            return DataType.CSV

        if url.path.endswith(".mdx") or url.path.endswith(".md"):
            logger.debug(f"Source of `{formatted_source}` detected as `mdx`.")
            return DataType.MDX

        if url.path.endswith(".docx"):
            logger.debug(f"Source of `{formatted_source}` detected as `docx`.")
            return DataType.DOCX

        if url.path.endswith(
            (".mp3", ".mp4", ".mp2", ".aac", ".wav", ".flac", ".pcm", ".m4a", ".ogg", ".opus", ".webm")
        ):
            logger.debug(f"Source of `{formatted_source}` detected as `audio`.")
            return DataType.AUDIO

        if url.path.endswith(".yaml"):
            try:
                response = requests.get(source)
                response.raise_for_status()
                try:
                    yaml_content = yaml.safe_load(response.text)
                except yaml.YAMLError as exc:
                    logger.error(f"Error parsing YAML: {exc}")
                    raise TypeError(f"Not a valid data type. Error loading YAML: {exc}")

                if is_openapi_yaml(yaml_content):
                    logger.debug(f"Source of `{formatted_source}` detected as `openapi`.")
                    return DataType.OPENAPI
                else:
                    logger.error(
                        f"Source of `{formatted_source}` does not contain all the required \
                        fields of OpenAPI yaml. Check 'https://spec.openapis.org/oas/v3.1.0'"
                    )
                    raise TypeError(
                        "Not a valid data type. Check 'https://spec.openapis.org/oas/v3.1.0', \
                        make sure you have all the required fields in YAML config data"
                    )
            except requests.exceptions.RequestException as e:
                logger.error(f"Error fetching URL {formatted_source}: {e}")

        if url.path.endswith(".json"):
            logger.debug(f"Source of `{formatted_source}` detected as `json_file`.")
            return DataType.JSON

        if "docs" in url.netloc or ("docs" in url.path and url.scheme != "file"):
            # `docs_site` detection via path is not accepted for local filesystem URIs,
            # because that would mean all paths that contain `docs` are now doc sites, which is too aggressive.
            logger.debug(f"Source of `{formatted_source}` detected as `docs_site`.")
            return DataType.DOCS_SITE

        if "github.com" in url.netloc:
            logger.debug(f"Source of `{formatted_source}` detected as `github`.")
            return DataType.GITHUB

        if is_google_drive_folder(url.netloc + url.path):
            logger.debug(f"Source of `{formatted_source}` detected as `google drive folder`.")
            return DataType.GOOGLE_DRIVE_FOLDER

        # If none of the above conditions are met, it's a general web page
        logger.debug(f"Source of `{formatted_source}` detected as `web_page`.")
        return DataType.WEB_PAGE

    elif not isinstance(source, str):
        # For datatypes where source is not a string.

        if isinstance(source, tuple) and len(source) == 2 and isinstance(source[0], str) and isinstance(source[1], str):
            logger.debug(f"Source of `{formatted_source}` detected as `qna_pair`.")
            return DataType.QNA_PAIR

        # Raise an error if it isn't a string and also not a valid non-string type (one of the previous).
        # We could stringify it, but it is better to raise an error and let the user decide how they want to do that.
        raise TypeError(
            "Source is not a string and a valid non-string type could not be detected. If you want to embed it, please stringify it, for instance by using `str(source)` or `(', ').join(source)`."  # noqa: E501
        )

    elif os.path.isfile(source):
        # For datatypes that support conventional file references.
        # Note: checking for string is not necessary anymore.

        if source.endswith(".docx"):
            logger.debug(f"Source of `{formatted_source}` detected as `docx`.")
            return DataType.DOCX

        if source.endswith(".csv"):
            logger.debug(f"Source of `{formatted_source}` detected as `csv`.")
            return DataType.CSV

        if source.endswith(".xml"):
            logger.debug(f"Source of `{formatted_source}` detected as `xml`.")
            return DataType.XML

        if source.endswith(".mdx") or source.endswith(".md"):
            logger.debug(f"Source of `{formatted_source}` detected as `mdx`.")
            return DataType.MDX

        if source.endswith(".txt"):
            logger.debug(f"Source of `{formatted_source}` detected as `text`.")
            return DataType.TEXT_FILE

        if source.endswith(".pdf"):
            logger.debug(f"Source of `{formatted_source}` detected as `pdf_file`.")
            return DataType.PDF_FILE

        if source.endswith(".yaml"):
            with open(source, "r") as file:
                yaml_content = yaml.safe_load(file)
                if is_openapi_yaml(yaml_content):
                    logger.debug(f"Source of `{formatted_source}` detected as `openapi`.")
                    return DataType.OPENAPI
                else:
                    logger.error(
                        f"Source of `{formatted_source}` does not contain all the required \
                                  fields of OpenAPI yaml. Check 'https://spec.openapis.org/oas/v3.1.0'"
                    )
                    raise ValueError(
                        "Invalid YAML data. Check 'https://spec.openapis.org/oas/v3.1.0', \
                        make sure to add all the required params"
                    )

        if source.endswith(".json"):
            logger.debug(f"Source of `{formatted_source}` detected as `json`.")
            return DataType.JSON

        if os.path.exists(source) and is_readable(open(source).read()):
            logger.debug(f"Source of `{formatted_source}` detected as `text_file`.")
            return DataType.TEXT_FILE

        # If the source is a valid file, that's not detectable as a type, an error is raised.
        # It does not fall back to text.
        raise ValueError(
            "Source points to a valid file, but based on the filename, no `data_type` can be detected. Please be aware, that not all data_types allow conventional file references, some require the use of the `file URI scheme`."  # noqa: E501
        )

    else:
        # Source is not a URL.

        # TODO: check if source is gmail query

        # check if the source is valid json string
        if is_valid_json_string(source):
            logger.debug(f"Source of `{formatted_source}` detected as `json`.")
            return DataType.JSON

        # Use text as final fallback.
        logger.debug(f"Source of `{formatted_source}` detected as `text`.")
        return DataType.TEXT


# check if the source is valid json string
def is_valid_json_string(source: str):
    try:
        _ = json.loads(source)
        return True
    except json.JSONDecodeError:
        return False


def validate_config(config_data):
    schema = Schema(
        {
            Optional("app"): {
                Optional("config"): {
                    Optional("id"): str,
                    Optional("name"): str,
                    Optional("log_level"): Or("DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"),
                    Optional("collect_metrics"): bool,
                    Optional("collection_name"): str,
                }
            },
            Optional("llm"): {
                Optional("provider"): Or(
                    "openai",
                    "azure_openai",
                    "anthropic",
                    "huggingface",
                    "cohere",
                    "together",
                    "gpt4all",
                    "ollama",
                    "jina",
                    "llama2",
                    "vertexai",
                    "google",
                    "aws_bedrock",
                    "mistralai",
                    "clarifai",
                    "vllm",
                    "groq",
                    "nvidia",
                ),
                Optional("config"): {
                    Optional("model"): str,
                    Optional("model_name"): str,
                    Optional("number_documents"): int,
                    Optional("temperature"): float,
                    Optional("max_tokens"): int,
                    Optional("top_p"): Or(float, int),
                    Optional("stream"): bool,
                    Optional("online"): bool,
                    Optional("token_usage"): bool,
                    Optional("template"): str,
                    Optional("prompt"): str,
                    Optional("system_prompt"): str,
                    Optional("deployment_name"): str,
                    Optional("where"): dict,
                    Optional("query_type"): str,
                    Optional("api_key"): str,
                    Optional("base_url"): str,
                    Optional("endpoint"): str,
                    Optional("model_kwargs"): dict,
                    Optional("local"): bool,
                    Optional("base_url"): str,
                    Optional("default_headers"): dict,
                    Optional("api_version"): Or(str, datetime.date),
                    Optional("http_client_proxies"): Or(str, dict),
                    Optional("http_async_client_proxies"): Or(str, dict),
                },
            },
            Optional("vectordb"): {
                Optional("provider"): Or(
                    "chroma", "elasticsearch", "opensearch", "lancedb", "pinecone", "qdrant", "weaviate", "zilliz"
                ),
                Optional("config"): object,  # TODO: add particular config schema for each provider
            },
            Optional("embedder"): {
                Optional("provider"): Or(
                    "openai",
                    "gpt4all",
                    "huggingface",
                    "vertexai",
                    "azure_openai",
                    "google",
                    "mistralai",
                    "clarifai",
                    "nvidia",
                    "ollama",
                    "cohere",
                    "aws_bedrock",
                ),
                Optional("config"): {
                    Optional("model"): Optional(str),
                    Optional("deployment_name"): Optional(str),
                    Optional("api_key"): str,
                    Optional("api_base"): str,
                    Optional("title"): str,
                    Optional("task_type"): str,
                    Optional("vector_dimension"): int,
                    Optional("base_url"): str,
                    Optional("endpoint"): str,
                    Optional("model_kwargs"): dict,
                    Optional("http_client_proxies"): Or(str, dict),
                    Optional("http_async_client_proxies"): Or(str, dict),
                },
            },
            Optional("embedding_model"): {
                Optional("provider"): Or(
                    "openai",
                    "gpt4all",
                    "huggingface",
                    "vertexai",
                    "azure_openai",
                    "google",
                    "mistralai",
                    "clarifai",
                    "nvidia",
                    "ollama",
                    "aws_bedrock",
                ),
                Optional("config"): {
                    Optional("model"): str,
                    Optional("deployment_name"): str,
                    Optional("api_key"): str,
                    Optional("title"): str,
                    Optional("task_type"): str,
                    Optional("vector_dimension"): int,
                    Optional("base_url"): str,
                },
            },
            Optional("chunker"): {
                Optional("chunk_size"): int,
                Optional("chunk_overlap"): int,
                Optional("length_function"): str,
                Optional("min_chunk_size"): int,
            },
            Optional("cache"): {
                Optional("similarity_evaluation"): {
                    Optional("strategy"): Or("distance", "exact"),
                    Optional("max_distance"): float,
                    Optional("positive"): bool,
                },
                Optional("config"): {
                    Optional("similarity_threshold"): float,
                    Optional("auto_flush"): int,
                },
            },
            Optional("memory"): {
                Optional("top_k"): int,
            },
        }
    )

    return schema.validate(config_data)


def chunks(iterable, batch_size=100, desc="Processing chunks"):
    """A helper function to break an iterable into chunks of size batch_size."""
    it = iter(iterable)
    total_size = len(iterable)

    with tqdm(total=total_size, desc=desc, unit="batch") as pbar:
        chunk = tuple(itertools.islice(it, batch_size))
        while chunk:
            yield chunk
            pbar.update(len(chunk))
            chunk = tuple(itertools.islice(it, batch_size))



import hashlib

from mem0x.bootstrap.helpers.json_serializable import register_deserializable
from mem0x.bootstrap.loaders.base_loader import BaseLoader


@register_deserializable
class LocalTextLoader(BaseLoader):
    def load_data(self, content):
        """Load data from a local text file."""
        url = "local"
        metadata = {
            "url": url,
        }
        doc_id = hashlib.sha256((content + url).encode()).hexdigest()
        return {
            "doc_id": doc_id,
            "data": [
                {
                    "content": content,
                    "meta_data": metadata,
                }
            ],
        }



import base64
import hashlib
import logging
import os
from email import message_from_bytes
from email.utils import parsedate_to_datetime
from textwrap import dedent
from typing import Optional

from bs4 import BeautifulSoup

try:
    from google.auth.transport.requests import Request
    from google.oauth2.credentials import Credentials
    from google_auth_oauthlib.flow import InstalledAppFlow
    from googleapiclient.discovery import build
except ImportError:
    raise ImportError("Gmail requires extra dependencies") from None

from mem0x.bootstrap.loaders.base_loader import BaseLoader
from mem0x.bootstrap.utils.misc import clean_string

logger = logging.getLogger(__name__)


class GmailReader:
    SCOPES = ["https://www.googleapis.com/auth/gmail.readonly"]

    def __init__(self, query: str, service=None, results_per_page: int = 10):
        self.query = query
        self.service = service or self._initialize_service()
        self.results_per_page = results_per_page

    @staticmethod
    def _initialize_service():
        credentials = GmailReader._get_credentials()
        return build("gmail", "v1", credentials=credentials)

    @staticmethod
    def _get_credentials():
        if not os.path.exists("credentials.json"):
            raise FileNotFoundError("Missing 'credentials.json'. Download it from your Google Developer account.")

        creds = (
            Credentials.from_authorized_user_file("token.json", GmailReader.SCOPES)
            if os.path.exists("token.json")
            else None
        )

        if not creds or not creds.valid:
            if creds and creds.expired and creds.refresh_token:
                creds.refresh(Request())
            else:
                flow = InstalledAppFlow.from_client_secrets_file("credentials.json", GmailReader.SCOPES)
                creds = flow.run_local_server(port=8080)
            with open("token.json", "w") as token:
                token.write(creds.to_json())
        return creds

    def load_emails(self) -> list[dict]:
        response = self.service.users().messages().list(userId="me", q=self.query).execute()
        messages = response.get("messages", [])

        return [self._parse_email(self._get_email(message["id"])) for message in messages]

    def _get_email(self, message_id: str):
        raw_message = self.service.users().messages().get(userId="me", id=message_id, format="raw").execute()
        return base64.urlsafe_b64decode(raw_message["raw"])

    def _parse_email(self, raw_email) -> dict:
        mime_msg = message_from_bytes(raw_email)
        return {
            "subject": self._get_header(mime_msg, "Subject"),
            "from": self._get_header(mime_msg, "From"),
            "to": self._get_header(mime_msg, "To"),
            "date": self._format_date(mime_msg),
            "body": self._get_body(mime_msg),
        }

    @staticmethod
    def _get_header(mime_msg, header_name: str) -> str:
        return mime_msg.get(header_name, "")

    @staticmethod
    def _format_date(mime_msg) -> Optional[str]:
        date_header = GmailReader._get_header(mime_msg, "Date")
        return parsedate_to_datetime(date_header).isoformat() if date_header else None

    @staticmethod
    def _get_body(mime_msg) -> str:
        def decode_payload(part):
            charset = part.get_content_charset() or "utf-8"
            try:
                return part.get_payload(decode=True).decode(charset)
            except UnicodeDecodeError:
                return part.get_payload(decode=True).decode(charset, errors="replace")

        if mime_msg.is_multipart():
            for part in mime_msg.walk():
                ctype = part.get_content_type()
                cdispo = str(part.get("Content-Disposition"))

                if ctype == "text/plain" and "attachment" not in cdispo:
                    return decode_payload(part)
                elif ctype == "text/html":
                    return decode_payload(part)
        else:
            return decode_payload(mime_msg)

        return ""


class GmailLoader(BaseLoader):
    def load_data(self, query: str):
        reader = GmailReader(query=query)
        emails = reader.load_emails()
        logger.info(f"Gmail Loader: {len(emails)} emails found for query '{query}'")

        data = []
        for email in emails:
            content = self._process_email(email)
            data.append({"content": content, "meta_data": email})

        return {"doc_id": self._generate_doc_id(query, data), "data": data}

    @staticmethod
    def _process_email(email: dict) -> str:
        content = BeautifulSoup(email["body"], "html.parser").get_text()
        content = clean_string(content)
        return dedent(
            f"""
            Email from '{email['from']}' to '{email['to']}'
            Subject: {email['subject']}
            Date: {email['date']}
            Content: {content}
        """
        )

    @staticmethod
    def _generate_doc_id(query: str, data: list[dict]) -> str:
        content_strings = [email["content"] for email in data]
        return hashlib.sha256((query + ", ".join(content_strings)).encode()).hexdigest()



import hashlib
import os

from mem0x.bootstrap.helpers.json_serializable import register_deserializable
from mem0x.bootstrap.loaders.base_loader import BaseLoader


@register_deserializable
class TextFileLoader(BaseLoader):
    def load_data(self, url: str):
        """Load data from a text file located at a local path."""
        if not os.path.exists(url):
            raise FileNotFoundError(f"The file at {url} does not exist.")

        with open(url, "r", encoding="utf-8") as file:
            content = file.read()

        doc_id = hashlib.sha256((content + url).encode()).hexdigest()

        metadata = {"url": url, "file_size": os.path.getsize(url), "file_type": url.split(".")[-1]}

        return {
            "doc_id": doc_id,
            "data": [
                {
                    "content": content,
                    "meta_data": metadata,
                }
            ],
        }



from typing import Any

from mem0x.bootstrap.helpers.json_serializable import JSONSerializable


class BaseConfig(JSONSerializable):
    """
    Base config.
    """

    def __init__(self):
        """Initializes a configuration class for a class."""
        pass

    def as_dict(self) -> dict[str, Any]:
        """Return config object as a dict

        :return: config object as dict
        :rtype: dict[str, Any]
        """
        return vars(self)



import os
from typing import Any, Dict, Optional

from pydantic import BaseModel, Field, model_validator


class PGVectorConfig(BaseModel):
    dbname: str = Field(os.getenv("POSTGRES_DBNAME") or "postgres", description="Default name for the database")
    collection_name: str = Field(
        os.getenv("POSTGRES_COLLECTION_NAME") or "mem0x", description="Default name for the collection"
    )
    embedding_model_dims: Optional[int] = Field(768, description="Dimensions of the embedding model")
    user: Optional[str] = Field(os.getenv("POSTGRES_USER") or "guest", description="Database user")
    password: Optional[str] = Field(os.getenv("POSTGRES_PASSWORD") or None, description="Database password")
    host: Optional[str] = Field(os.getenv("POSTGRES_HOST") or None, description="Database host. Default is localhost")
    port: Optional[int] = Field(5432, description="Database port. Default is 5432")
    diskann: Optional[bool] = Field(True, description="Use diskann for approximate nearest neighbors search")
    hnsw: Optional[bool] = Field(False, description="Use hnsw for faster search")

    @model_validator(mode="before")
    @classmethod
    def validate_extra_fields(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        allowed_fields = set(cls.model_fields.keys())
        input_fields = set(values.keys())
        extra_fields = input_fields - allowed_fields
        if extra_fields:
            raise ValueError(
                f"Extra fields not allowed: {', '.join(extra_fields)}. Please input only the following fields: {', '.join(allowed_fields)}"
            )
        return values



from importlib import import_module
from typing import Any, Optional

from mem0x.bootstrap.chunkers.base_chunker import BaseChunker
from mem0x.bootstrap.config import AddConfig
from mem0x.bootstrap.config.add_config import ChunkerConfig, LoaderConfig
from mem0x.bootstrap.helpers.json_serializable import JSONSerializable
from mem0x.bootstrap.loaders.base_loader import BaseLoader
from mem0x.bootstrap.models.data_type import DataType


class DataFormatter(JSONSerializable):
    """
    DataFormatter is an internal utility class which abstracts the mapping for
    loaders and chunkers to the data_type entered by the user in their
    .add or .add_local method call
    """

    def __init__(
        self,
        data_type: DataType,
        config: AddConfig,
        loader: Optional[BaseLoader] = None,
        chunker: Optional[BaseChunker] = None,
    ):
        """
        Initialize a dataformatter, set data type and chunker based on datatype.

        :param data_type: The type of the data to load and chunk.
        :type data_type: DataType
        :param config: AddConfig instance with nested loader and chunker config attributes.
        :type config: AddConfig
        """
        self.loader = self._get_loader(data_type=data_type, config=config.loader, loader=loader)
        self.chunker = self._get_chunker(data_type=data_type, config=config.chunker, chunker=chunker)

    @staticmethod
    def _lazy_load(module_path: str):
        module_path, class_name = module_path.rsplit(".", 1)
        module = import_module(module_path)
        return getattr(module, class_name)

    def _get_loader(
        self,
        data_type: DataType,
        config: LoaderConfig,
        loader: Optional[BaseLoader],
        **kwargs: Optional[dict[str, Any]],
    ) -> BaseLoader:
        """
        Returns the appropriate data loader for the given data type.

        :param data_type: The type of the data to load.
        :type data_type: DataType
        :param config: Config to initialize the loader with.
        :type config: LoaderConfig
        :raises ValueError: If an unsupported data type is provided.
        :return: The loader for the given data type.
        :rtype: BaseLoader
        """
        loaders = {
            DataType.YOUTUBE_VIDEO: "mem0x.bootstrap.loaders.youtube_video.YoutubeVideoLoader",
            DataType.PDF_FILE: "mem0x.bootstrap.loaders.pdf_file.PdfFileLoader",
            DataType.WEB_PAGE: "mem0x.bootstrap.loaders.web_page.WebPageLoader",
            DataType.QNA_PAIR: "mem0x.bootstrap.loaders.local_qna_pair.LocalQnaPairLoader",
            DataType.TEXT: "mem0x.bootstrap.loaders.local_text.LocalTextLoader",
            DataType.DOCX: "mem0x.bootstrap.loaders.docx_file.DocxFileLoader",
            DataType.SITEMAP: "mem0x.bootstrap.loaders.sitemap.SitemapLoader",
            DataType.XML: "mem0x.bootstrap.loaders.xml.XmlLoader",
            DataType.DOCS_SITE: "mem0x.bootstrap.loaders.docs_site_loader.DocsSiteLoader",
            DataType.CSV: "mem0x.bootstrap.loaders.csv.CsvLoader",
            DataType.MDX: "mem0x.bootstrap.loaders.mdx.MdxLoader",
            DataType.IMAGE: "mem0x.bootstrap.loaders.image.ImageLoader",
            DataType.UNSTRUCTURED: "mem0x.bootstrap.loaders.unstructured_file.UnstructuredLoader",
            DataType.JSON: "mem0x.bootstrap.loaders.json.JSONLoader",
            DataType.OPENAPI: "mem0x.bootstrap.loaders.openapi.OpenAPILoader",
            DataType.GMAIL: "mem0x.bootstrap.loaders.gmail.GmailLoader",
            DataType.NOTION: "mem0x.bootstrap.loaders.notion.NotionLoader",
            DataType.SUBSTACK: "mem0x.bootstrap.loaders.substack.SubstackLoader",
            DataType.YOUTUBE_CHANNEL: "mem0x.bootstrap.loaders.youtube_channel.YoutubeChannelLoader",
            DataType.DISCORD: "mem0x.bootstrap.loaders.discord.DiscordLoader",
            DataType.RSSFEED: "mem0x.bootstrap.loaders.rss_feed.RSSFeedLoader",
            DataType.BEEHIIV: "mem0x.bootstrap.loaders.beehiiv.BeehiivLoader",
            DataType.GOOGLE_DRIVE: "mem0x.bootstrap.loaders.google_drive.GoogleDriveLoader",
            DataType.DIRECTORY: "mem0x.bootstrap.loaders.directory_loader.DirectoryLoader",
            DataType.SLACK: "mem0x.bootstrap.loaders.slack.SlackLoader",
            DataType.DROPBOX: "mem0x.bootstrap.loaders.dropbox.DropboxLoader",
            DataType.TEXT_FILE: "mem0x.bootstrap.loaders.text_file.TextFileLoader",
            DataType.EXCEL_FILE: "mem0x.bootstrap.loaders.excel_file.ExcelFileLoader",
            DataType.AUDIO: "mem0x.bootstrap.loaders.audio.AudioLoader",
        }

        if data_type == DataType.CUSTOM or loader is not None:
            loader_class: type = loader
            if loader_class:
                return loader_class
        elif data_type in loaders:
            loader_class: type = self._lazy_load(loaders[data_type])
            return loader_class()

        raise ValueError(
            f"Cant find the loader for {data_type}.\
                    We recommend to pass the loader to use data_type: {data_type}"
        )

    def _get_chunker(self, data_type: DataType, config: ChunkerConfig, chunker: Optional[BaseChunker]) -> BaseChunker:
        """Returns the appropriate chunker for the given data type (updated for lazy loading)."""
        chunker_classes = {
            DataType.YOUTUBE_VIDEO: "mem0x.bootstrap.chunkers.youtube_video.YoutubeVideoChunker",
            DataType.PDF_FILE: "mem0x.bootstrap.chunkers.pdf_file.PdfFileChunker",
            DataType.WEB_PAGE: "mem0x.bootstrap.chunkers.web_page.WebPageChunker",
            DataType.QNA_PAIR: "mem0x.bootstrap.chunkers.qna_pair.QnaPairChunker",
            DataType.TEXT: "mem0x.bootstrap.chunkers.text.TextChunker",
            DataType.DOCX: "mem0x.bootstrap.chunkers.docx_file.DocxFileChunker",
            DataType.SITEMAP: "mem0x.bootstrap.chunkers.sitemap.SitemapChunker",
            DataType.XML: "mem0x.bootstrap.chunkers.xml.XmlChunker",
            DataType.DOCS_SITE: "mem0x.bootstrap.chunkers.docs_site.DocsSiteChunker",
            DataType.CSV: "mem0x.bootstrap.chunkers.table.TableChunker",
            DataType.MDX: "mem0x.bootstrap.chunkers.mdx.MdxChunker",
            DataType.IMAGE: "mem0x.bootstrap.chunkers.image.ImageChunker",
            DataType.UNSTRUCTURED: "mem0x.bootstrap.chunkers.unstructured_file.UnstructuredFileChunker",
            DataType.JSON: "mem0x.bootstrap.chunkers.json.JSONChunker",
            DataType.OPENAPI: "mem0x.bootstrap.chunkers.openapi.OpenAPIChunker",
            DataType.GMAIL: "mem0x.bootstrap.chunkers.gmail.GmailChunker",
            DataType.NOTION: "mem0x.bootstrap.chunkers.notion.NotionChunker",
            DataType.SUBSTACK: "mem0x.bootstrap.chunkers.substack.SubstackChunker",
            DataType.YOUTUBE_CHANNEL: "mem0x.bootstrap.chunkers.common_chunker.CommonChunker",
            DataType.DISCORD: "mem0x.bootstrap.chunkers.common_chunker.CommonChunker",
            DataType.CUSTOM: "mem0x.bootstrap.chunkers.common_chunker.CommonChunker",
            DataType.RSSFEED: "mem0x.bootstrap.chunkers.rss_feed.RSSFeedChunker",
            DataType.BEEHIIV: "mem0x.bootstrap.chunkers.beehiiv.BeehiivChunker",
            DataType.GOOGLE_DRIVE: "mem0x.bootstrap.chunkers.google_drive.GoogleDriveChunker",
            DataType.DIRECTORY: "mem0x.bootstrap.chunkers.common_chunker.CommonChunker",
            DataType.SLACK: "mem0x.bootstrap.chunkers.common_chunker.CommonChunker",
            DataType.DROPBOX: "mem0x.bootstrap.chunkers.common_chunker.CommonChunker",
            DataType.TEXT_FILE: "mem0x.bootstrap.chunkers.common_chunker.CommonChunker",
            DataType.EXCEL_FILE: "mem0x.bootstrap.chunkers.excel_file.ExcelFileChunker",
            DataType.AUDIO: "mem0x.bootstrap.chunkers.audio.AudioChunker",
        }

        if chunker is not None:
            return chunker
        elif data_type in chunker_classes:
            chunker_class = self._lazy_load(chunker_classes[data_type])
            chunker = chunker_class(config)
            chunker.set_data_type(data_type)
            return chunker

        raise ValueError(
            f"Cant find the chunker for {data_type}.\
                We recommend to pass the chunker to use data_type: {data_type}"
        )



import hashlib
import importlib.util

try:
    import unstructured  # noqa: F401
    from langchain_community.document_loaders import UnstructuredExcelLoader
except ImportError:
    raise ImportError(
        'Excel file requires extra dependencies. Install with `pip install "unstructured[local-inference, all-docs]"`'
    ) from None

if importlib.util.find_spec("openpyxl") is None and importlib.util.find_spec("xlrd") is None:
    raise ImportError("Excel file requires extra dependencies. Install with `pip install openpyxl xlrd`") from None

from mem0x.bootstrap.helpers.json_serializable import register_deserializable
from mem0x.bootstrap.loaders.base_loader import BaseLoader
from mem0x.bootstrap.utils.misc import clean_string


@register_deserializable
class ExcelFileLoader(BaseLoader):
    def load_data(self, excel_url):
        """Load data from a Excel file."""
        loader = UnstructuredExcelLoader(excel_url)
        pages = loader.load_and_split()

        data = []
        for page in pages:
            content = page.page_content
            content = clean_string(content)

            metadata = page.metadata
            metadata["url"] = excel_url

            data.append({"content": content, "meta_data": metadata})

        doc_id = hashlib.sha256((content + excel_url).encode()).hexdigest()
        return {
            "doc_id": doc_id,
            "data": data,
        }



from datetime import datetime, timedelta

import requests

URL_SHORTTERM_GET_RUNIDS = "http://127.0.0.1:19527/shortterm_get_runids"
URL_SHORTTERM_GET = "http://127.0.0.1:19527/shortterm_get"
URL_SHORTTERM_DELETE = "http://127.0.0.1:19527/shortterm_delete"
URL_LONGTERM_ADD = "http://127.0.0.1:19527/longterm_add"


def get_runids(comparison_time):
    data = {"timestamp": comparison_time.isoformat()}
    try:
        response = requests.post(URL_SHORTTERM_GET_RUNIDS, json=data)

        if response.status_code == 200:
            return response.json()["data"]
        else:
            print("Failed with status code:", response.status_code)
            return []
    except Exception as e:
        print(f"An error occurred: {e}")
        return []


def get_memory(run_id, comparison_time):
    data = {"run_id": run_id, "timestamp": comparison_time.isoformat()}
    try:
        response = requests.post(URL_SHORTTERM_GET, json=data)
        if response.status_code == 200:
            return response.json()["data"]
        else:
            print("Failed with status code:", response.status_code)
            return []
    except Exception as e:
        print(f"An error occurred: {e}")
        return []


def add_longterm(run_id, conversation):
    try:
        messages = []
        user_id = ""
        agent_id = ""
        for item in conversation:
            role = "user"
            text = item["content"]["text"]
            if item["user_id"] == item["agent_id"]:
                role = "assistant"
                agent_id = item["agent_id"]
            else:
                role = "user"
                user_id = item["user_id"]
            messages.append({"role": role, "content": text})
        data = {"messages": messages, "user_id": user_id, "agent_id": agent_id, "run_id": run_id}
        response = requests.post(URL_LONGTERM_ADD, json=data)
        if response.status_code == 200:
            return True
        else:
            print("Failed with status code:", response.status_code)
            return False
    except Exception as e:
        print(f"An error occurred: {e}")
        return False


def delete_shortterm(id):
    data = {"id": id}
    try:
        response = requests.post(URL_SHORTTERM_DELETE, json=data)
        if response.status_code == 200:
            return
        else:
            print("Failed with status code:", response.status_code)
            return
    except Exception as e:
        print(f"An error occurred: {e}")
        return


def main():
    comparison_time = datetime.now() - timedelta(hours=0)
    runids = get_runids(comparison_time)
    for item1 in runids:
        conversation = get_memory(item1["run_id"], comparison_time)
        print(conversation)
        if add_longterm(item1["run_id"], conversation):
            for item2 in conversation:
                delete_shortterm(item2["id"])


if __name__ == "__main__":
    main()



from enum import Enum


class DirectDataType(Enum):
    """
    DirectDataType enum contains data types that contain raw data directly.
    """

    TEXT = "text"


class IndirectDataType(Enum):
    """
    IndirectDataType enum contains data types that contain references to data stored elsewhere.
    """

    YOUTUBE_VIDEO = "youtube_video"
    PDF_FILE = "pdf_file"
    WEB_PAGE = "web_page"
    SITEMAP = "sitemap"
    XML = "xml"
    DOCX = "docx"
    DOCS_SITE = "docs_site"
    NOTION = "notion"
    CSV = "csv"
    MDX = "mdx"
    IMAGE = "image"
    UNSTRUCTURED = "unstructured"
    JSON = "json"
    OPENAPI = "openapi"
    GMAIL = "gmail"
    SUBSTACK = "substack"
    YOUTUBE_CHANNEL = "youtube_channel"
    DISCORD = "discord"
    CUSTOM = "custom"
    RSSFEED = "rss_feed"
    BEEHIIV = "beehiiv"
    GOOGLE_DRIVE = "google_drive"
    DIRECTORY = "directory"
    SLACK = "slack"
    DROPBOX = "dropbox"
    TEXT_FILE = "text_file"
    EXCEL_FILE = "excel_file"
    AUDIO = "audio"


class SpecialDataType(Enum):
    """
    SpecialDataType enum contains data types that are neither direct nor indirect, or simply require special attention.
    """

    QNA_PAIR = "qna_pair"


class DataType(Enum):
    TEXT = DirectDataType.TEXT.value
    YOUTUBE_VIDEO = IndirectDataType.YOUTUBE_VIDEO.value
    PDF_FILE = IndirectDataType.PDF_FILE.value
    WEB_PAGE = IndirectDataType.WEB_PAGE.value
    SITEMAP = IndirectDataType.SITEMAP.value
    XML = IndirectDataType.XML.value
    DOCX = IndirectDataType.DOCX.value
    DOCS_SITE = IndirectDataType.DOCS_SITE.value
    NOTION = IndirectDataType.NOTION.value
    CSV = IndirectDataType.CSV.value
    MDX = IndirectDataType.MDX.value
    QNA_PAIR = SpecialDataType.QNA_PAIR.value
    IMAGE = IndirectDataType.IMAGE.value
    UNSTRUCTURED = IndirectDataType.UNSTRUCTURED.value
    JSON = IndirectDataType.JSON.value
    OPENAPI = IndirectDataType.OPENAPI.value
    GMAIL = IndirectDataType.GMAIL.value
    SUBSTACK = IndirectDataType.SUBSTACK.value
    YOUTUBE_CHANNEL = IndirectDataType.YOUTUBE_CHANNEL.value
    DISCORD = IndirectDataType.DISCORD.value
    CUSTOM = IndirectDataType.CUSTOM.value
    RSSFEED = IndirectDataType.RSSFEED.value
    BEEHIIV = IndirectDataType.BEEHIIV.value
    GOOGLE_DRIVE = IndirectDataType.GOOGLE_DRIVE.value
    DIRECTORY = IndirectDataType.DIRECTORY.value
    SLACK = IndirectDataType.SLACK.value
    DROPBOX = IndirectDataType.DROPBOX.value
    TEXT_FILE = IndirectDataType.TEXT_FILE.value
    EXCEL_FILE = IndirectDataType.EXCEL_FILE.value
    AUDIO = IndirectDataType.AUDIO.value



import hashlib
import os

import validators

from mem0x.bootstrap.helpers.json_serializable import register_deserializable
from mem0x.bootstrap.loaders.base_loader import BaseLoader

try:
    from deepgram import DeepgramClient, PrerecordedOptions
except ImportError:
    raise ImportError(
        "Audio file requires extra dependencies. Install with `pip install deepgram-sdk==3.2.7`"
    ) from None


@register_deserializable
class AudioLoader(BaseLoader):
    def __init__(self):
        if not os.environ.get("DEEPGRAM_API_KEY"):
            raise ValueError("DEEPGRAM_API_KEY is not set")

        DG_KEY = os.environ.get("DEEPGRAM_API_KEY")
        self.client = DeepgramClient(DG_KEY)

    def load_data(self, url: str):
        """Load data from a audio file or URL."""

        options = PrerecordedOptions(
            model="nova-2",
            smart_format=True,
        )
        if validators.url(url):
            source = {"url": url}
            response = self.client.listen.prerecorded.v("1").transcribe_url(source, options)
        else:
            with open(url, "rb") as audio:
                source = {"buffer": audio}
                response = self.client.listen.prerecorded.v("1").transcribe_file(source, options)
        content = response["results"]["channels"][0]["alternatives"][0]["transcript"]

        doc_id = hashlib.sha256((content + url).encode()).hexdigest()
        metadata = {"url": url}

        return {
            "doc_id": doc_id,
            "data": [
                {
                    "content": content,
                    "meta_data": metadata,
                }
            ],
        }



import hashlib
import logging
from typing import Any, Optional

from mem0x.bootstrap.loaders.base_loader import BaseLoader

logger = logging.getLogger(__name__)


class PostgresLoader(BaseLoader):
    def __init__(self, config: Optional[dict[str, Any]] = None):
        super().__init__()
        if not config:
            raise ValueError(f"Must provide the valid config. Received: {config}")

        self.connection = None
        self.cursor = None
        self._setup_loader(config=config)

    def _setup_loader(self, config: dict[str, Any]):
        try:
            import psycopg
        except ImportError as e:
            raise ImportError("Unable to import required packages") from e

        if "url" in config:
            config_info = config.get("url")
        else:
            conn_params = []
            for key, value in config.items():
                conn_params.append(f"{key}={value}")
            config_info = " ".join(conn_params)

        logger.info(f"Connecting to postrgres sql: {config_info}")
        self.connection = psycopg.connect(conninfo=config_info)
        self.cursor = self.connection.cursor()

    @staticmethod
    def _check_query(query):
        if not isinstance(query, str):
            raise ValueError(
                f"Invalid postgres query: {query}. Provide the valid source to add from postgres",  # noqa:E501
            )

    def load_data(self, query):
        self._check_query(query)
        try:
            data = []
            data_content = []
            self.cursor.execute(query)
            results = self.cursor.fetchall()
            for result in results:
                doc_content = str(result)
                data.append({"content": doc_content, "meta_data": {"url": query}})
                data_content.append(doc_content)
            doc_id = hashlib.sha256((query + ", ".join(data_content)).encode()).hexdigest()
            return {
                "doc_id": doc_id,
                "data": data,
            }
        except Exception as e:
            raise ValueError(f"Failed to load data using query={query} with: {e}")

    def close_connection(self):
        if self.cursor:
            self.cursor.close()
            self.cursor = None
        if self.connection:
            self.connection.close()
            self.connection = None



import json
import sqlite3
import uuid
from datetime import datetime


class SQLiteManager:
    def __init__(self, db_path=":memory:"):
        self.connection = sqlite3.connect(db_path, check_same_thread=False)
        self._create_shortterm_memories_table()

    def _create_shortterm_memories_table(self):
        with self.connection:
            self.connection.execute(
                """
                CREATE TABLE IF NOT EXISTS shortterm_memories (
                  id TEXT PRIMARY KEY,
                  content TEXT,
                  user_id TEXT,
                  agent_id TEXT,
                  run_id TEXT,
                  table_name TEXT,
                  created_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP
                )
            """
            )
            self.connection.execute(
                """
                CREATE INDEX IF NOT EXISTS idx_shortterm_memories on shortterm_memories("user_id", "agent_id", "run_id", "created_at")
            """
            )

    def add(
        self,
        content,
        user_id,
        agent_id,
        run_id,
        table_name,
        created_at=None,
    ):
        with self.connection:
            self.connection.execute(
                """
                INSERT INTO shortterm_memories (id, content, user_id, agent_id, run_id, table_name, created_at)
                VALUES (?, ?, ?, ?, ?, ?, ?)
            """,
                (
                    str(uuid.uuid4()),
                    content.json(),
                    user_id,
                    agent_id,
                    run_id,
                    table_name,
                    created_at or datetime.now(),
                ),
            )

    def get(self, table_name, run_id, timestamp):
        cursor = self.connection.execute(
            """
            SELECT id, content, user_id, agent_id, run_id, created_at
            FROM shortterm_memories
            WHERE run_id = ? and table_name = ? and created_at <= ?
            ORDER BY created_at ASC
        """,
            (
                run_id,
                table_name,
                timestamp,
            ),
        )
        rows = cursor.fetchall()
        return [
            {
                "id": row[0],
                "content": json.loads(row[1]),
                "user_id": row[2],
                "agent_id": row[3],
                "run_id": row[4],
                "created_at": row[5],
            }
            for row in rows
        ]

    def get_runids(self, timestamp):
        cursor = self.connection.execute(
            """
            SELECT DISTINCT run_id
            FROM shortterm_memories where created_at <= ?
        """,
            (timestamp,),
        )
        rows = cursor.fetchall()
        return [
            {
                "run_id": row[0],
            }
            for row in rows
        ]

    def delete(self, id):
        with self.connection:
            self.connection.execute(
                """DELETE from shortterm_memories where id = ?""",
                (id,),
            )

    def reset(self):
        with self.connection:
            self.connection.execute("DROP INDEX IF EXISTS idx_shortterm_memories")
            self.connection.execute("DROP TABLE IF EXISTS shortterm_memories")
            self._create_shortterm_memories_table()



import hashlib
from io import StringIO
from urllib.parse import urlparse

import requests
import yaml

from mem0x.bootstrap.loaders.base_loader import BaseLoader


class OpenAPILoader(BaseLoader):
    @staticmethod
    def _get_file_content(content):
        url = urlparse(content)
        if all([url.scheme, url.netloc]) and url.scheme not in ["file", "http", "https"]:
            raise ValueError("Not a valid URL.")

        if url.scheme in ["http", "https"]:
            response = requests.get(content)
            response.raise_for_status()
            return StringIO(response.text)
        elif url.scheme == "file":
            path = url.path
            return open(path)
        else:
            return open(content)

    @staticmethod
    def load_data(content):
        """Load yaml file of openapi. Each pair is a document."""
        data = []
        file_path = content
        data_content = []
        with OpenAPILoader._get_file_content(content=content) as file:
            yaml_data = yaml.load(file, Loader=yaml.SafeLoader)
            for i, (key, value) in enumerate(yaml_data.items()):
                string_data = f"{key}: {value}"
                metadata = {"url": file_path, "row": i + 1}
                data.append({"content": string_data, "meta_data": metadata})
                data_content.append(string_data)
        doc_id = hashlib.sha256((content + ", ".join(data_content)).encode()).hexdigest()
        return {"doc_id": doc_id, "data": data}



from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel


class LongTermSearchRequest(BaseModel):
    table_name: str
    query: str
    user_id: Optional[str] = None
    agent_id: Optional[str] = None
    run_id: Optional[str] = None
    limit: Optional[int] = 100
    filters: Optional[str] = None


class Data(BaseModel):
    role: str
    content: str


class LongTermAddRequest(BaseModel):
    table_name: str
    messages: Union[str, List[Data]]
    user_id: Optional[str] = None
    agent_id: Optional[str] = None
    run_id: Optional[str] = None
    metadata: Optional[str] = None
    filters: Optional[str] = None


class LongTermRemoveRequest(BaseModel):
    memory_id: str


class LongTermRemoveAllRequest(BaseModel):
    user_id: Optional[str] = None
    agent_id: Optional[str] = None
    run_id: Optional[str] = None


class LongTermUpdateRequest(BaseModel):
    memory_id: str
    data: Dict


class Content(BaseModel):
    text: str
    action: Optional[str] = None


class ShortTermAddRequest(BaseModel):
    content: Content
    user_id: str
    agent_id: Optional[str] = None
    run_id: str
    table_name: str


class ShortTermGetRequest(BaseModel):
    run_id: str
    timestamp: Optional[str] = None
    table_name: str


class ShortTermGetRunIdsRequest(BaseModel):
    timestamp: Optional[str] = None


class ShortTermDeleteRequest(BaseModel):
    id: str


class BootstrapRequest(BaseModel):
    data_type: str
    src: str
    table_name: str
    chunk_size: int = 2000
    chunk_overlap: int = 50
    min_chunk_size: int = 50
    user_id: Optional[str] = None
    agent_id: Optional[str] = None
    run_id: Optional[str] = None


class Response(BaseModel):
    message: str
    status: bool
    data: Optional[Any] = None



import hashlib

try:
    import unstructured  # noqa: F401
    from langchain_community.document_loaders import UnstructuredXMLLoader
except ImportError:
    raise ImportError(
        'XML file requires extra dependencies. Install with `pip install "unstructured[local-inference, all-docs]"`'
    ) from None
from mem0x.bootstrap.helpers.json_serializable import register_deserializable
from mem0x.bootstrap.loaders.base_loader import BaseLoader
from mem0x.bootstrap.utils.misc import clean_string


@register_deserializable
class XmlLoader(BaseLoader):
    def load_data(self, xml_url):
        """Load data from a XML file."""
        loader = UnstructuredXMLLoader(xml_url)
        data = loader.load()
        content = data[0].page_content
        content = clean_string(content)
        metadata = data[0].metadata
        metadata["url"] = metadata["source"]
        del metadata["source"]
        output = [{"content": content, "meta_data": metadata}]
        doc_id = hashlib.sha256((content + xml_url).encode()).hexdigest()
        return {
            "doc_id": doc_id,
            "data": output,
        }



import hashlib

from mem0x.bootstrap.helpers.json_serializable import register_deserializable
from mem0x.bootstrap.loaders.base_loader import BaseLoader
from mem0x.bootstrap.utils.misc import clean_string


@register_deserializable
class UnstructuredLoader(BaseLoader):
    def load_data(self, url):
        """Load data from an Unstructured file."""
        try:
            import unstructured  # noqa: F401
            from langchain_community.document_loaders import \
                UnstructuredFileLoader
        except ImportError:
            raise ImportError(
                'Unstructured file requires extra dependencies. Install with `pip install "unstructured[local-inference, all-docs]"`'  # noqa: E501
            ) from None

        loader = UnstructuredFileLoader(url)
        data = []
        all_content = []
        pages = loader.load_and_split()
        if not len(pages):
            raise ValueError("No data found")
        for page in pages:
            content = page.page_content
            content = clean_string(content)
            metadata = page.metadata
            metadata["url"] = url
            data.append(
                {
                    "content": content,
                    "meta_data": metadata,
                }
            )
            all_content.append(content)
        doc_id = hashlib.sha256((" ".join(all_content) + url).encode()).hexdigest()
        return {
            "doc_id": doc_id,
            "data": data,
        }



import hashlib
import re

try:
    from googleapiclient.errors import HttpError
except ImportError:
    raise ImportError("Google Drive requires extra dependencies.") from None

from langchain_community.document_loaders import GoogleDriveLoader as Loader

try:
    import unstructured  # noqa: F401
    from langchain_community.document_loaders import UnstructuredFileIOLoader
except ImportError:
    raise ImportError(
        'Unstructured file requires extra dependencies. Install with `pip install "unstructured[local-inference, all-docs]"`'  # noqa: E501
    ) from None

from mem0x.bootstrap.helpers.json_serializable import register_deserializable
from mem0x.bootstrap.loaders.base_loader import BaseLoader


@register_deserializable
class GoogleDriveLoader(BaseLoader):
    @staticmethod
    def _get_drive_id_from_url(url: str):
        regex = r"^https:\/\/drive\.google\.com\/drive\/(?:u\/\d+\/)folders\/([a-zA-Z0-9_-]+)$"
        if re.match(regex, url):
            return url.split("/")[-1]
        raise ValueError(
            f"The url provided {url} does not match a google drive folder url. Example drive url: "
            f"https://drive.google.com/drive/u/0/folders/xxxx"
        )

    def load_data(self, url: str):
        """Load data from a Google drive folder."""
        folder_id: str = self._get_drive_id_from_url(url)

        try:
            loader = Loader(
                folder_id=folder_id,
                recursive=True,
                file_loader_cls=UnstructuredFileIOLoader,
            )

            data = []
            all_content = []

            docs = loader.load()
            for doc in docs:
                all_content.append(doc.page_content)
                # renames source to url for later use.
                doc.metadata["url"] = doc.metadata.pop("source")
                data.append({"content": doc.page_content, "meta_data": doc.metadata})

            doc_id = hashlib.sha256((" ".join(all_content) + url).encode()).hexdigest()
            return {"doc_id": doc_id, "data": data}

        except HttpError:
            raise FileNotFoundError("Unable to locate folder or files, check provided drive URL and try again")



import hashlib
import logging
from typing import Any, Optional

from mem0x.bootstrap.loaders.base_loader import BaseLoader
from mem0x.bootstrap.utils.misc import clean_string

logger = logging.getLogger(__name__)


class MySQLLoader(BaseLoader):
    def __init__(self, config: Optional[dict[str, Any]]):
        super().__init__()
        if not config:
            raise ValueError(
                f"Invalid sql config: {config}.",
                "Provide the correct config.",
            )

        self.config = config
        self.connection = None
        self.cursor = None
        self._setup_loader(config=config)

    def _setup_loader(self, config: dict[str, Any]):
        try:
            import mysql.connector as sqlconnector
        except ImportError as e:
            raise ImportError(
                "Unable to import required packages for MySQL loader."  # noqa: E501
            ) from e

        try:
            self.connection = sqlconnector.connection.MySQLConnection(**config)
            self.cursor = self.connection.cursor()
        except (sqlconnector.Error, IOError) as err:
            logger.info(f"Connection failed: {err}")
            raise ValueError(
                f"Unable to connect with the given config: {config}.",
                "Please provide the correct configuration to load data from you MySQL DB.",
            )

    @staticmethod
    def _check_query(query):
        if not isinstance(query, str):
            raise ValueError(
                f"Invalid mysql query: {query}",
                "Provide the valid query to add from mysql, \
                    make sure you are following",
            )

    def load_data(self, query):
        self._check_query(query=query)
        data = []
        data_content = []
        self.cursor.execute(query)
        rows = self.cursor.fetchall()
        for row in rows:
            doc_content = clean_string(str(row))
            data.append({"content": doc_content, "meta_data": {"url": query}})
            data_content.append(doc_content)
        doc_id = hashlib.sha256((query + ", ".join(data_content)).encode()).hexdigest()
        return {
            "doc_id": doc_id,
            "data": data,
        }



import concurrent.futures
import hashlib
import logging
import re
import shlex
from typing import Any, Optional

from tqdm import tqdm

from mem0x.bootstrap.loaders.base_loader import BaseLoader
from mem0x.bootstrap.utils.misc import clean_string

GITHUB_URL = "https://github.com"
GITHUB_API_URL = "https://api.github.com"

VALID_SEARCH_TYPES = set(["code", "repo", "pr", "issue", "discussion", "branch", "file"])


class GithubLoader(BaseLoader):
    """Load data from GitHub search query."""

    def __init__(self, config: Optional[dict[str, Any]] = None):
        super().__init__()
        if not config:
            raise ValueError(
                "GithubLoader requires a personal access token to use github api. Check - `https://docs.github.com/en/authentication/keeping-your-account-and-data-secure/managing-your-personal-access-tokens#creating-a-personal-access-token-classic`"  # noqa: E501
            )

        try:
            from github import Github
        except ImportError as e:
            raise ValueError(
                "GithubLoader requires extra dependencies. \
                  Install with `pip install gitpython==3.1.38 PyGithub==1.59.1`"
            ) from e

        self.config = config
        token = config.get("token")
        if not token:
            raise ValueError(
                "GithubLoader requires a personal access token to use github api. Check - `https://docs.github.com/en/authentication/keeping-your-account-and-data-secure/managing-your-personal-access-tokens#creating-a-personal-access-token-classic`"  # noqa: E501
            )

        try:
            self.client = Github(token)
        except Exception as e:
            logging.error(f"GithubLoader failed to initialize client: {e}")
            self.client = None

    def _github_search_code(self, query: str):
        """Search GitHub code."""
        data = []
        results = self.client.search_code(query)
        for result in tqdm(results, total=results.totalCount, desc="Loading code files from github"):
            url = result.html_url
            logging.info(f"Added data from url: {url}")
            content = result.decoded_content.decode("utf-8")
            metadata = {
                "url": url,
            }
            data.append(
                {
                    "content": clean_string(content),
                    "meta_data": metadata,
                }
            )
        return data

    def _get_github_repo_data(self, repo_name: str, branch_name: str = None, file_path: str = None) -> list[dict]:
        """Get file contents from Repo"""
        data = []

        repo = self.client.get_repo(repo_name)
        repo_contents = repo.get_contents("")

        if branch_name:
            repo_contents = repo.get_contents("", ref=branch_name)
        if file_path:
            repo_contents = [repo.get_contents(file_path)]

        with tqdm(desc="Loading files:", unit="item") as progress_bar:
            while repo_contents:
                file_content = repo_contents.pop(0)
                if file_content.type == "dir":
                    try:
                        repo_contents.extend(repo.get_contents(file_content.path))
                    except Exception:
                        logging.warning(f"Failed to read directory: {file_content.path}")
                        progress_bar.update(1)
                        continue
                else:
                    try:
                        file_text = file_content.decoded_content.decode()
                    except Exception:
                        logging.warning(f"Failed to read file: {file_content.path}")
                        progress_bar.update(1)
                        continue

                    file_path = file_content.path
                    data.append(
                        {
                            "content": clean_string(file_text),
                            "meta_data": {
                                "path": file_path,
                            },
                        }
                    )

                progress_bar.update(1)

        return data

    def _github_search_repo(self, query: str) -> list[dict]:
        """Search GitHub repo."""

        logging.info(f"Searching github repos with query: {query}")
        updated_query = query.split(":")[-1]
        data = self._get_github_repo_data(updated_query)
        return data

    def _github_search_issues_and_pr(self, query: str, type: str) -> list[dict]:
        """Search GitHub issues and PRs."""
        data = []

        query = f"{query} is:{type}"
        logging.info(f"Searching github for query: {query}")

        results = self.client.search_issues(query)

        logging.info(f"Total results: {results.totalCount}")
        for result in tqdm(results, total=results.totalCount, desc=f"Loading {type} from github"):
            url = result.html_url
            title = result.title
            body = result.body
            if not body:
                logging.warning(f"Skipping issue because empty content for: {url}")
                continue
            labels = " ".join([label.name for label in result.labels])
            issue_comments = result.get_comments()
            comments = []
            comments_created_at = []
            for comment in issue_comments:
                comments_created_at.append(str(comment.created_at))
                comments.append(f"{comment.user.name}:{comment.body}")
            content = "\n".join([title, labels, body, *comments])
            metadata = {
                "url": url,
                "created_at": str(result.created_at),
                "comments_created_at": " ".join(comments_created_at),
            }
            data.append(
                {
                    "content": clean_string(content),
                    "meta_data": metadata,
                }
            )
        return data

    # need to test more for discussion
    def _github_search_discussions(self, query: str):
        """Search GitHub discussions."""
        data = []

        query = f"{query} is:discussion"
        logging.info(f"Searching github repo for query: {query}")
        repos_results = self.client.search_repositories(query)
        logging.info(f"Total repos found: {repos_results.totalCount}")
        for repo_result in tqdm(repos_results, total=repos_results.totalCount, desc="Loading discussions from github"):
            teams = repo_result.get_teams()
            for team in teams:
                team_discussions = team.get_discussions()
                for discussion in team_discussions:
                    url = discussion.html_url
                    title = discussion.title
                    body = discussion.body
                    if not body:
                        logging.warning(f"Skipping discussion because empty content for: {url}")
                        continue
                    comments = []
                    comments_created_at = []
                    print("Discussion comments: ", discussion.comments_url)
                    content = "\n".join([title, body, *comments])
                    metadata = {
                        "url": url,
                        "created_at": str(discussion.created_at),
                        "comments_created_at": " ".join(comments_created_at),
                    }
                    data.append(
                        {
                            "content": clean_string(content),
                            "meta_data": metadata,
                        }
                    )
        return data

    def _get_github_repo_branch(self, query: str, type: str) -> list[dict]:
        """Get file contents for specific branch"""

        logging.info(f"Searching github repo for query: {query} is:{type}")
        pattern = r"repo:(\S+) name:(\S+)"
        match = re.search(pattern, query)

        if match:
            repo_name = match.group(1)
            branch_name = match.group(2)
        else:
            raise ValueError(
                f"Repository name and Branch name not found, instead found this \
                    Repo: {repo_name}, Branch: {branch_name}"
            )

        data = self._get_github_repo_data(repo_name=repo_name, branch_name=branch_name)
        return data

    def _get_github_repo_file(self, query: str, type: str) -> list[dict]:
        """Get specific file content"""

        logging.info(f"Searching github repo for query: {query} is:{type}")
        pattern = r"repo:(\S+) path:(\S+)"
        match = re.search(pattern, query)

        if match:
            repo_name = match.group(1)
            file_path = match.group(2)
        else:
            raise ValueError(
                f"Repository name and File name not found, instead found this Repo: {repo_name}, File: {file_path}"
            )

        data = self._get_github_repo_data(repo_name=repo_name, file_path=file_path)
        return data

    def _search_github_data(self, search_type: str, query: str):
        """Search github data."""
        if search_type == "code":
            data = self._github_search_code(query)
        elif search_type == "repo":
            data = self._github_search_repo(query)
        elif search_type == "issue":
            data = self._github_search_issues_and_pr(query, search_type)
        elif search_type == "pr":
            data = self._github_search_issues_and_pr(query, search_type)
        elif search_type == "branch":
            data = self._get_github_repo_branch(query, search_type)
        elif search_type == "file":
            data = self._get_github_repo_file(query, search_type)
        elif search_type == "discussion":
            raise ValueError("GithubLoader does not support searching discussions yet.")
        else:
            raise NotImplementedError(f"{search_type} not supported")

        return data

    @staticmethod
    def _get_valid_github_query(query: str):
        """Check if query is valid and return search types and valid GitHub query."""
        query_terms = shlex.split(query)
        # query must provide repo to load data from
        if len(query_terms) < 1 or "repo:" not in query:
            raise ValueError(
                "GithubLoader requires a search query with `repo:` term."  # noqa: E501
            )

        github_query = []
        types = set()
        type_pattern = r"type:([a-zA-Z,]+)"
        for term in query_terms:
            term_match = re.search(type_pattern, term)
            if term_match:
                search_types = term_match.group(1).split(",")
                types.update(search_types)
            else:
                github_query.append(term)

        # query must provide search type
        if len(types) == 0:
            raise ValueError(
                "GithubLoader requires a search query with `type:` term."  # noqa: E501
            )

        for search_type in search_types:
            if search_type not in VALID_SEARCH_TYPES:
                raise ValueError(
                    f"Invalid search type: {search_type}. Valid types are: {', '.join(VALID_SEARCH_TYPES)}"
                )

        query = " ".join(github_query)

        return types, query

    def load_data(self, search_query: str, max_results: int = 1000):
        """Load data from GitHub search query."""

        if not self.client:
            raise ValueError(
                "GithubLoader client is not initialized, data will not be loaded."  # noqa: E501
            )

        search_types, query = self._get_valid_github_query(search_query)
        logging.info(f"Searching github for query: {query}, with types: {', '.join(search_types)}")

        data = []

        with concurrent.futures.ThreadPoolExecutor(max_workers=4) as executor:
            futures_map = executor.map(self._search_github_data, search_types, [query] * len(search_types))
            for search_data in tqdm(futures_map, total=len(search_types), desc="Searching data from github"):
                data.extend(search_data)

        return {
            "doc_id": hashlib.sha256(query.encode()).hexdigest(),
            "data": data,
        }



import hashlib
import logging
from typing import Any, Optional

from mem0x.bootstrap.config.add_config import ChunkerConfig
from mem0x.bootstrap.helpers.json_serializable import JSONSerializable
from mem0x.bootstrap.models.data_type import DataType

logger = logging.getLogger(__name__)


class BaseChunker(JSONSerializable):
    def __init__(self, text_splitter):
        """Initialize the chunker."""
        self.text_splitter = text_splitter
        self.data_type = None

    def create_chunks(
        self,
        loader,
        src,
        app_id=None,
        config: Optional[ChunkerConfig] = None,
        **kwargs: Optional[dict[str, Any]],
    ):
        """
        Loads data and chunks it.

        :param loader: The loader whose `load_data` method is used to create
        the raw data.
        :param src: The data to be handled by the loader. Can be a URL for
        remote sources or local content for local loaders.
        :param app_id: App id used to generate the doc_id.
        """
        documents = []
        chunk_ids = []
        id_map = {}
        min_chunk_size = config.min_chunk_size if config is not None else 1
        logger.info(f"Skipping chunks smaller than {min_chunk_size} characters")
        data_result = loader.load_data(src, **kwargs)
        data_records = data_result["data"]
        doc_id = data_result["doc_id"]
        # Prefix app_id in the document id if app_id is not None to
        # distinguish between different documents stored in the same
        # elasticsearch or opensearch index
        doc_id = f"{app_id}--{doc_id}" if app_id is not None else doc_id
        metadatas = []
        for data in data_records:
            content = data["content"]

            metadata = data["meta_data"]
            # add data type to meta data to allow query using data type
            metadata["data_type"] = self.data_type.value
            metadata["doc_id"] = doc_id

            # TODO: Currently defaulting to the src as the url. This is done intentianally since some
            # of the data types like 'gmail' loader doesn't have the url in the meta data.
            url = metadata.get("url", src)

            chunks = self.get_chunks(content)
            for chunk in chunks:
                chunk_id = hashlib.sha256((chunk + url).encode()).hexdigest()
                chunk_id = f"{app_id}--{chunk_id}" if app_id is not None else chunk_id
                if id_map.get(chunk_id) is None and len(chunk) >= min_chunk_size:
                    id_map[chunk_id] = True
                    chunk_ids.append(chunk_id)
                    documents.append(chunk)
                    metadatas.append(metadata)
        return {
            "documents": documents,
            "ids": chunk_ids,
            "metadatas": metadatas,
            "doc_id": doc_id,
        }

    def get_chunks(self, content):
        """
        Returns chunks using text splitter instance.

        Override in child class if custom logic.
        """
        return self.text_splitter.split_text(content)

    def set_data_type(self, data_type: DataType):
        """
        set the data type of chunker
        """
        self.data_type = data_type

        # TODO: This should be done during initialization. This means it has to be done in the child classes.

    @staticmethod
    def get_word_count(documents) -> int:
        return sum(len(document.split(" ")) for document in documents)



import json
import logging
from string import Template
from typing import Any, Type, TypeVar, Union

T = TypeVar("T", bound="JSONSerializable")

# NOTE: Through inheritance, all of our classes should be children of JSONSerializable. (highest level)
# NOTE: The @register_deserializable decorator should be added to all user facing child classes. (lowest level)

logger = logging.getLogger(__name__)


def register_deserializable(cls: Type[T]) -> Type[T]:
    """
    A class decorator to register a class as deserializable.

    When a class is decorated with @register_deserializable, it becomes
    a part of the set of classes that the JSONSerializable class can
    deserialize.

    Deserialization is in essence loading attributes from a json file.
    This decorator is a security measure put in place to make sure that
    you don't load attributes that were initially part of another class.

    Example:
        @register_deserializable
        class ChildClass(JSONSerializable):
            def __init__(self, ...):
                # initialization logic

    Args:
        cls (Type): The class to be registered.

    Returns:
        Type: The same class, after registration.
    """
    JSONSerializable._register_class_as_deserializable(cls)
    return cls


class JSONSerializable:
    """
    A class to represent a JSON serializable object.

    This class provides methods to serialize and deserialize objects,
    as well as to save serialized objects to a file and load them back.
    """

    _deserializable_classes = set()  # Contains classes that are whitelisted for deserialization.

    def serialize(self) -> str:
        """
        Serialize the object to a JSON-formatted string.

        Returns:
            str: A JSON string representation of the object.
        """
        try:
            return json.dumps(self, default=self._auto_encoder, ensure_ascii=False)
        except Exception as e:
            logger.error(f"Serialization error: {e}")
            return "{}"

    @classmethod
    def deserialize(cls, json_str: str) -> Any:
        """
        Deserialize a JSON-formatted string to an object.
        If it fails, a default class is returned instead.
        Note: This *returns* an instance, it's not automatically loaded on the calling class.

        Example:
            app = App.deserialize(json_str)

        Args:
            json_str (str): A JSON string representation of an object.

        Returns:
            Object: The deserialized object.
        """
        try:
            return json.loads(json_str, object_hook=cls._auto_decoder)
        except Exception as e:
            logger.error(f"Deserialization error: {e}")
            # Return a default instance in case of failure
            return cls()

    @staticmethod
    def _auto_encoder(obj: Any) -> Union[dict[str, Any], None]:
        """
        Automatically encode an object for JSON serialization.

        Args:
            obj (Object): The object to be encoded.

        Returns:
            dict: A dictionary representation of the object.
        """
        if hasattr(obj, "__dict__"):
            dct = {}
            for key, value in obj.__dict__.items():
                try:
                    # Recursive: If the value is an instance of a subclass of JSONSerializable,
                    # serialize it using the JSONSerializable serialize method.
                    if isinstance(value, JSONSerializable):
                        serialized_value = value.serialize()
                        # The value is stored as a serialized string.
                        dct[key] = json.loads(serialized_value)
                    # Custom rules (subclass is not json serializable by default)
                    elif isinstance(value, Template):
                        dct[key] = {"__type__": "Template", "data": value.template}
                    # Future custom types we can follow a similar pattern
                    # elif isinstance(value, SomeOtherType):
                    #     dct[key] = {
                    #         "__type__": "SomeOtherType",
                    #         "data": value.some_method()
                    #     }
                    # NOTE: Keep in mind that this logic needs to be applied to the decoder too.
                    else:
                        json.dumps(value)  # Try to serialize the value.
                        dct[key] = value
                except TypeError:
                    pass  # If it fails, simply pass to skip this key-value pair of the dictionary.

            dct["__class__"] = obj.__class__.__name__
            return dct
        raise TypeError(f"Object of type {type(obj)} is not JSON serializable")

    @classmethod
    def _auto_decoder(cls, dct: dict[str, Any]) -> Any:
        """
        Automatically decode a dictionary to an object during JSON deserialization.

        Args:
            dct (dict): The dictionary representation of an object.

        Returns:
            Object: The decoded object or the original dictionary if decoding is not possible.
        """
        class_name = dct.pop("__class__", None)
        if class_name:
            if not hasattr(cls, "_deserializable_classes"):  # Additional safety check
                raise AttributeError(f"`{class_name}` has no registry of allowed deserializations.")
            if class_name not in {cl.__name__ for cl in cls._deserializable_classes}:
                raise KeyError(f"Deserialization of class `{class_name}` is not allowed.")
            target_class = next((cl for cl in cls._deserializable_classes if cl.__name__ == class_name), None)
            if target_class:
                obj = target_class.__new__(target_class)
                for key, value in dct.items():
                    if isinstance(value, dict) and "__type__" in value:
                        if value["__type__"] == "Template":
                            value = Template(value["data"])
                        # For future custom types we can follow a similar pattern
                        # elif value["__type__"] == "SomeOtherType":
                        #     value = SomeOtherType.some_constructor(value["data"])
                    default_value = getattr(target_class, key, None)
                    setattr(obj, key, value or default_value)
                return obj
        return dct

    def save_to_file(self, filename: str) -> None:
        """
        Save the serialized object to a file.

        Args:
            filename (str): The path to the file where the object should be saved.
        """
        with open(filename, "w", encoding="utf-8") as f:
            f.write(self.serialize())

    @classmethod
    def load_from_file(cls, filename: str) -> Any:
        """
        Load and deserialize an object from a file.

        Args:
            filename (str): The path to the file from which the object should be loaded.

        Returns:
            Object: The deserialized object.
        """
        with open(filename, "r", encoding="utf-8") as f:
            json_str = f.read()
            return cls.deserialize(json_str)

    @classmethod
    def _register_class_as_deserializable(cls, target_class: Type[T]) -> None:
        """
        Register a class as deserializable. This is a classmethod and globally shared.

        This method adds the target class to the set of classes that
        can be deserialized. This is a security measure to ensure only
        whitelisted classes are deserialized.

        Args:
            target_class (Type): The class to be registered.
        """
        cls._deserializable_classes.add(target_class)



import hashlib

try:
    from langchain_community.document_loaders import Docx2txtLoader
except ImportError:
    raise ImportError("Docx file requires extra dependencies. Install with `pip install docx2txt==0.8`") from None
from mem0x.bootstrap.helpers.json_serializable import register_deserializable
from mem0x.bootstrap.loaders.base_loader import BaseLoader


@register_deserializable
class DocxFileLoader(BaseLoader):
    def load_data(self, url):
        """Load data from a .docx file."""
        loader = Docx2txtLoader(url)
        output = []
        data = loader.load()
        content = data[0].page_content
        metadata = data[0].metadata
        metadata["url"] = "local"
        output.append({"content": content, "meta_data": metadata})
        doc_id = hashlib.sha256((content + url).encode()).hexdigest()
        return {
            "doc_id": doc_id,
            "data": output,
        }



import logging

import typer
import uvicorn

app = typer.Typer()


def setup_logging(level: int):
    # setup logging
    logging.basicConfig(
        level=level,
        format="[%(asctime)s] [%(thread)d] [%(levelname)s] %(message)s",
        filename="app.log",
        filemode="a",
    )
    console_handler = logging.StreamHandler()
    console_handler.setLevel(level)
    formatter = logging.Formatter("[%(asctime)s] [%(thread)d] [%(levelname)s] %(message)s")
    console_handler.setFormatter(formatter)
    logging.getLogger("").addHandler(console_handler)
    logging.debug("logging start...")


@app.command()
def serve(
    host: str = "127.0.0.1",
    port: int = 19527,
    workers: int = 1,
):
    setup_logging(logging.DEBUG)
    uvicorn.run("mem0x.app.serve:app", host=host, port=port, workers=workers, reload=False)


def run():
    app()



from typing import Any, Optional

from mem0x.bootstrap.helpers.json_serializable import JSONSerializable


class BaseLoader(JSONSerializable):
    def __init__(self):
        pass

    def load_data(self, url, **kwargs: Optional[dict[str, Any]]):
        """
        Implemented by child classes
        """
        pass



import uuid
from datetime import datetime
from typing import Dict, List, Optional

import psycopg2
# from psycopg2 import sql
from psycopg2.extras import RealDictCursor


class PostgreSQLManager:
    def __init__(self, dbname: str, user: str, password: str, host: str = "localhost", port: int = 5432):
        """Initialize the PostgreSQLManager with a connection to the database.

        Args:
            dbname (str): Name of the PostgreSQL database.
            user (str): Database user.
            password (str): Database password.
            host (str): Database host. Defaults to "localhost".
            port (int): Database port. Defaults to 5432.
        """
        self.connection = psycopg2.connect(dbname=dbname, user=user, password=password, host=host, port=port)
        self._create_history_table()

    def _create_history_table(self) -> None:
        """Create the history table if it does not exist."""
        with self.connection.cursor() as cursor:
            cursor.execute("""
                CREATE TABLE IF NOT EXISTS history (
                    id UUID PRIMARY KEY,
                    memory_id TEXT,
                    old_memory TEXT,
                    new_memory TEXT,
                    new_value TEXT,
                    event TEXT,
                    created_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
                    updated_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
                    is_deleted INTEGER
                );
            """)
            self.connection.commit()
            cursor.close()

    def add_history(
        self,
        memory_id: str,
        old_memory: str,
        new_memory: str,
        event: str,
        created_at: Optional[datetime] = None,
        updated_at: Optional[datetime] = None,
        is_deleted: int = 0,
    ) -> None:
        """Add a new history entry to the database.

        Args:
            memory_id (str): The ID of the memory being tracked.
            old_memory (str): The previous state of the memory.
            new_memory (str): The new state of the memory.
            event (str): The event that triggered the change.
            created_at (Optional[datetime]): The creation timestamp. Defaults to current timestamp.
            updated_at (Optional[datetime]): The update timestamp. Defaults to current timestamp.
            is_deleted (int): Flag indicating if the entry is deleted. Defaults to 0.
        """
        with self.connection.cursor() as cursor:
            cursor.execute(
                """
                INSERT INTO history (id, memory_id, old_memory, new_memory, event, created_at, updated_at, is_deleted)
                VALUES (%s, %s, %s, %s, %s, %s, %s, %s);
            """,
                (
                    str(uuid.uuid4()),
                    memory_id,
                    old_memory,
                    new_memory,
                    event,
                    created_at or datetime.now(),
                    updated_at or datetime.now(),
                    is_deleted,
                ),
            )
            self.connection.commit()
            cursor.close()

    def get_history(self, memory_id: str) -> List[Dict[str, str]]:
        """Retrieve the history entries for a specific memory ID.

        Args:
            memory_id (str): The ID of the memory to retrieve history for.

        Returns:
            List[Dict[str, str]]: A list of history entries as dictionaries.
        """
        with self.connection.cursor(cursor_factory=RealDictCursor) as cursor:
            cursor.execute(
                """
                SELECT id, memory_id, old_memory, new_memory, event, created_at, updated_at
                FROM history
                WHERE memory_id = %s
                ORDER BY updated_at ASC;
            """,
                (memory_id,),
            )
            self.connection.commit()
            rows = cursor.fetchall()
            cursor.close()
            return rows

    def reset(self) -> None:
        """Reset the history table by dropping and recreating it."""
        with self.connection.cursor() as cursor:
            cursor.execute("DROP TABLE IF EXISTS history;")
            self.connection.commit()
            cursor.close()
            self._create_history_table()

    def close(self) -> None:
        """Close the database connection."""
        self.connection.close()



import hashlib
import logging
import os
import ssl
from typing import Any, Optional

import certifi

from mem0x.bootstrap.loaders.base_loader import BaseLoader
from mem0x.bootstrap.utils.misc import clean_string

SLACK_API_BASE_URL = "https://www.slack.com/api/"

logger = logging.getLogger(__name__)


class SlackLoader(BaseLoader):
    def __init__(self, config: Optional[dict[str, Any]] = None):
        super().__init__()

        self.config = config if config else {}

        if "base_url" not in self.config:
            self.config["base_url"] = SLACK_API_BASE_URL

        self.client = None
        self._setup_loader(self.config)

    def _setup_loader(self, config: dict[str, Any]):
        try:
            from slack_sdk import WebClient
        except ImportError as e:
            raise ImportError(
                "Slack loader requires extra dependencies. \
                Install with `pip install --upgrade mem0x[slack]`"
            ) from e

        if os.getenv("SLACK_USER_TOKEN") is None:
            raise ValueError(
                "SLACK_USER_TOKEN environment variables not provided."  # noqa:E501
            )

        logger.info(f"Creating Slack Loader with config: {config}")
        # get slack client config params
        slack_bot_token = os.getenv("SLACK_USER_TOKEN")
        ssl_cert = ssl.create_default_context(cafile=certifi.where())
        base_url = config.get("base_url", SLACK_API_BASE_URL)
        headers = config.get("headers")
        # for Org-Wide App
        team_id = config.get("team_id")

        self.client = WebClient(
            token=slack_bot_token,
            base_url=base_url,
            ssl=ssl_cert,
            headers=headers,
            team_id=team_id,
        )
        logger.info("Slack Loader setup successful!")

    @staticmethod
    def _check_query(query):
        if not isinstance(query, str):
            raise ValueError(
                f"Invalid query passed to Slack loader, found: {query}."  # noqa:E501
            )

    def load_data(self, query):
        self._check_query(query)
        try:
            data = []
            data_content = []

            logger.info(f"Searching slack conversations for query: {query}")
            results = self.client.search_messages(
                query=query,
                sort="timestamp",
                sort_dir="desc",
                count=self.config.get("count", 100),
            )

            messages = results.get("messages")
            num_message = len(messages)
            logger.info(f"Found {num_message} messages for query: {query}")

            matches = messages.get("matches", [])
            for message in matches:
                url = message.get("permalink")
                text = message.get("text")
                content = clean_string(text)

                message_meta_data_keys = ["iid", "team", "ts", "type", "user", "username"]
                metadata = {}
                for key in message.keys():
                    if key in message_meta_data_keys:
                        metadata[key] = message.get(key)
                metadata.update({"url": url})

                data.append(
                    {
                        "content": content,
                        "meta_data": metadata,
                    }
                )
                data_content.append(content)
            doc_id = hashlib.md5((query + ", ".join(data_content)).encode()).hexdigest()
            return {
                "doc_id": doc_id,
                "data": data,
            }
        except Exception as e:
            logger.warning(f"Error in loading slack data: {e}")
            raise ValueError(
                f"Error in loading slack data: {e}."  # noqa:E501
            ) from e



import logging

from mem0x.configs.base import ShortTermMemoryConfig
from mem0x.memory.base import ShortTermMemoryBase
from mem0x.memory.shortterm_storage import SQLiteManager
from mem0x.memory.shortterm_storage_postgres import PostgreSQLManager

logger = logging.getLogger(__name__)


class ShortTermMemory(ShortTermMemoryBase):
    def __init__(self, config: ShortTermMemoryConfig = ShortTermMemoryConfig()):
        self.config = config

        if self.config.postgres_storage_config:
            self.db = PostgreSQLManager(
                self.config.postgres_storage_config.dbname,
                self.config.postgres_storage_config.user,
                self.config.postgres_storage_config.password,
                self.config.postgres_storage_config.host,
                self.config.postgres_storage_config.port,
            )
        else:
            self.db = SQLiteManager(self.config.sqlite_db_path)

    def add(self, content, user_id, agent_id, run_id, table_name, created_at=None):
        self.db.add(content, user_id, agent_id, run_id, table_name, created_at)

    def get(self, table_name, run_id, timestamp):
        """
        Retrieve a memory by ID.

        Args:
            memory_id (str): ID of the memory to retrieve.

        Returns:
            List: Retrieved memory.
        """
        result = self.db.get(table_name, run_id, timestamp)
        return result

    def get_runids(self, timestamp):
        return self.db.get_runids(timestamp)

    def delete(self, id):
        self.db.delete(id)

    def reset(self):
        self.db.reset()



import hashlib
import os

from dropbox.files import FileMetadata

from mem0x.bootstrap.helpers.json_serializable import register_deserializable
from mem0x.bootstrap.loaders.base_loader import BaseLoader
from mem0x.bootstrap.loaders.directory_loader import DirectoryLoader


@register_deserializable
class DropboxLoader(BaseLoader):
    def __init__(self):
        access_token = os.environ.get("DROPBOX_ACCESS_TOKEN")
        if not access_token:
            raise ValueError("Please set the `DROPBOX_ACCESS_TOKEN` environment variable.")
        try:
            from dropbox import Dropbox, exceptions
        except ImportError:
            raise ImportError("Dropbox requires extra dependencies. Install with `pip install dropbox==11.36.2`")

        try:
            dbx = Dropbox(access_token)
            dbx.users_get_current_account()
            self.dbx = dbx
        except exceptions.AuthError as ex:
            raise ValueError("Invalid Dropbox access token. Please verify your token and try again.") from ex

    def _download_folder(self, path: str, local_root: str) -> list[FileMetadata]:
        """Download a folder from Dropbox and save it preserving the directory structure."""
        entries = self.dbx.files_list_folder(path).entries
        for entry in entries:
            local_path = os.path.join(local_root, entry.name)
            if isinstance(entry, FileMetadata):
                self.dbx.files_download_to_file(local_path, f"{path}/{entry.name}")
            else:
                os.makedirs(local_path, exist_ok=True)
                self._download_folder(f"{path}/{entry.name}", local_path)
        return entries

    def _generate_dir_id_from_all_paths(self, path: str) -> str:
        """Generate a unique ID for a directory based on all of its paths."""
        entries = self.dbx.files_list_folder(path).entries
        paths = [f"{path}/{entry.name}" for entry in entries]
        return hashlib.sha256("".join(paths).encode()).hexdigest()

    def load_data(self, path: str):
        """Load data from a Dropbox URL, preserving the folder structure."""
        root_dir = f"dropbox_{self._generate_dir_id_from_all_paths(path)}"
        os.makedirs(root_dir, exist_ok=True)

        for entry in self.dbx.files_list_folder(path).entries:
            local_path = os.path.join(root_dir, entry.name)
            if isinstance(entry, FileMetadata):
                self.dbx.files_download_to_file(local_path, f"{path}/{entry.name}")
            else:
                os.makedirs(local_path, exist_ok=True)
                self._download_folder(f"{path}/{entry.name}", local_path)

        dir_loader = DirectoryLoader()
        data = dir_loader.load_data(root_dir)["data"]

        # Clean up
        self._clean_directory(root_dir)

        return {
            "doc_id": hashlib.sha256(path.encode()).hexdigest(),
            "data": data,
        }

    def _clean_directory(self, dir_path):
        """Recursively delete a directory and its contents."""
        for item in os.listdir(dir_path):
            item_path = os.path.join(dir_path, item)
            if os.path.isdir(item_path):
                self._clean_directory(item_path)
            else:
                os.remove(item_path)
        os.rmdir(dir_path)



import hashlib
import logging
from typing import Any, Optional

import requests

try:
    from bs4 import BeautifulSoup
except ImportError:
    raise ImportError(
        "Webpage requires extra dependencies. Install with `pip install beautifulsoup4==4.12.3`"
    ) from None

from mem0x.bootstrap.helpers.json_serializable import register_deserializable
from mem0x.bootstrap.loaders.base_loader import BaseLoader
from mem0x.bootstrap.utils.misc import clean_string

logger = logging.getLogger(__name__)


@register_deserializable
class WebPageLoader(BaseLoader):
    # Shared session for all instances
    _session = requests.Session()

    def load_data(self, url, **kwargs: Optional[dict[str, Any]]):
        """Load data from a web page using a shared requests' session."""
        all_references = False
        for key, value in kwargs.items():
            if key == "all_references":
                all_references = kwargs["all_references"]
        headers = {
            "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/98.0.4758.102 Safari/537.36",  # noqa:E501
        }
        response = self._session.get(url, headers=headers, timeout=30)
        response.raise_for_status()
        data = response.content
        reference_links = self.fetch_reference_links(response)
        if all_references:
            for i in reference_links:
                try:
                    response = self._session.get(i, headers=headers, timeout=30)
                    response.raise_for_status()
                    data += response.content
                except Exception as e:
                    logging.error(f"Failed to add URL {url}: {e}")
                    continue

        content = self._get_clean_content(data, url)

        metadata = {"url": url}

        doc_id = hashlib.sha256((content + url).encode()).hexdigest()
        return {
            "doc_id": doc_id,
            "data": [
                {
                    "content": content,
                    "meta_data": metadata,
                }
            ],
        }

    @staticmethod
    def _get_clean_content(html, url) -> str:
        soup = BeautifulSoup(html, "html.parser")
        original_size = len(str(soup.get_text()))

        tags_to_exclude = [
            "nav",
            "aside",
            "form",
            "header",
            "noscript",
            "svg",
            "canvas",
            "footer",
            "script",
            "style",
        ]
        for tag in soup(tags_to_exclude):
            tag.decompose()

        ids_to_exclude = ["sidebar", "main-navigation", "menu-main-menu"]
        for id_ in ids_to_exclude:
            tags = soup.find_all(id=id_)
            for tag in tags:
                tag.decompose()

        classes_to_exclude = [
            "elementor-location-header",
            "navbar-header",
            "nav",
            "header-sidebar-wrapper",
            "blog-sidebar-wrapper",
            "related-posts",
        ]
        for class_name in classes_to_exclude:
            tags = soup.find_all(class_=class_name)
            for tag in tags:
                tag.decompose()

        content = soup.get_text()
        content = clean_string(content)

        cleaned_size = len(content)
        if original_size != 0:
            logger.info(
                f"[{url}] Cleaned page size: {cleaned_size} characters, down from {original_size} (shrunk: {original_size-cleaned_size} chars, {round((1-(cleaned_size/original_size)) * 100, 2)}%)"  # noqa:E501
            )

        return content

    @classmethod
    def close_session(cls):
        cls._session.close()

    def fetch_reference_links(self, response):
        if response.status_code == 200:
            soup = BeautifulSoup(response.content, "html.parser")
            a_tags = soup.find_all("a", href=True)
            reference_links = [a["href"] for a in a_tags if a["href"].startswith("http")]
            return reference_links
        else:
            print(f"Failed to retrieve the page. Status code: {response.status_code}")
            return []



import concurrent.futures
import hashlib
import logging
import os
from urllib.parse import urlparse

import requests
from tqdm import tqdm

try:
    from bs4 import BeautifulSoup
    from bs4.builder import ParserRejectedMarkup
except ImportError:
    raise ImportError(
        "Sitemap requires extra dependencies. Install with `pip install beautifulsoup4==4.12.3`"
    ) from None

from mem0x.bootstrap.helpers.json_serializable import register_deserializable
from mem0x.bootstrap.loaders.base_loader import BaseLoader
from mem0x.bootstrap.loaders.web_page import WebPageLoader

logger = logging.getLogger(__name__)


@register_deserializable
class SitemapLoader(BaseLoader):
    """
    This method takes a sitemap URL or local file path as input and retrieves
    all the URLs to use the WebPageLoader to load content
    of each page.
    """

    def load_data(self, sitemap_source):
        output = []
        web_page_loader = WebPageLoader()
        headers = {
            "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/98.0.4758.102 Safari/537.36",  # noqa:E501
        }

        if urlparse(sitemap_source).scheme in ("http", "https"):
            try:
                response = requests.get(sitemap_source, headers=headers)
                response.raise_for_status()
                soup = BeautifulSoup(response.text, "xml")
            except requests.RequestException as e:
                logger.error(f"Error fetching sitemap from URL: {e}")
                return
        elif os.path.isfile(sitemap_source):
            with open(sitemap_source, "r") as file:
                soup = BeautifulSoup(file, "xml")
        else:
            raise ValueError("Invalid sitemap source. Please provide a valid URL or local file path.")

        links = [link.text for link in soup.find_all("loc") if link.parent.name == "url"]
        if len(links) == 0:
            links = [link.text for link in soup.find_all("loc")]

        doc_id = hashlib.sha256((" ".join(links) + sitemap_source).encode()).hexdigest()

        def load_web_page(link):
            try:
                loader_data = web_page_loader.load_data(link)
                return loader_data.get("data")
            except ParserRejectedMarkup as e:
                logger.error(f"Failed to parse {link}: {e}")
            return None

        with concurrent.futures.ThreadPoolExecutor() as executor:
            future_to_link = {executor.submit(load_web_page, link): link for link in links}
            for future in tqdm(concurrent.futures.as_completed(future_to_link), total=len(links), desc="Loading pages"):
                link = future_to_link[future]
                try:
                    data = future.result()
                    if data:
                        output.extend(data)
                except Exception as e:
                    logger.error(f"Error loading page {link}: {e}")

        return {"doc_id": doc_id, "data": output}



import concurrent.futures
import hashlib
import logging

from tqdm import tqdm

from mem0x.bootstrap.loaders.base_loader import BaseLoader
from mem0x.bootstrap.loaders.youtube_video import YoutubeVideoLoader

logger = logging.getLogger(__name__)


class YoutubeChannelLoader(BaseLoader):
    """Loader for youtube channel."""

    def load_data(self, channel_name):
        try:
            import yt_dlp
        except ImportError as e:
            raise ValueError(
                "YoutubeChannelLoader requires extra dependencies. Install with `pip install yt_dlp==2023.11.14 youtube-transcript-api==0.6.1`"  # noqa: E501
            ) from e

        data = []
        data_urls = []
        youtube_url = f"https://www.youtube.com/{channel_name}/videos"
        youtube_video_loader = YoutubeVideoLoader()

        def _get_yt_video_links():
            try:
                ydl_opts = {
                    "quiet": True,
                    "extract_flat": True,
                }
                with yt_dlp.YoutubeDL(ydl_opts) as ydl:
                    info_dict = ydl.extract_info(youtube_url, download=False)
                    if "entries" in info_dict:
                        videos = [entry["url"] for entry in info_dict["entries"]]
                        return videos
            except Exception:
                logger.error(f"Failed to fetch youtube videos for channel: {channel_name}")
                return []

        def _load_yt_video(video_link):
            try:
                each_load_data = youtube_video_loader.load_data(video_link)
                if each_load_data:
                    return each_load_data.get("data")
            except Exception as e:
                logger.error(f"Failed to load youtube video {video_link}: {e}")
            return None

        def _add_youtube_channel():
            video_links = _get_yt_video_links()
            logger.info("Loading videos from youtube channel...")
            with concurrent.futures.ThreadPoolExecutor() as executor:
                # Submitting all tasks and storing the future object with the video link
                future_to_video = {
                    executor.submit(_load_yt_video, video_link): video_link for video_link in video_links
                }

                for future in tqdm(
                    concurrent.futures.as_completed(future_to_video), total=len(video_links), desc="Processing videos"
                ):
                    video = future_to_video[future]
                    try:
                        results = future.result()
                        if results:
                            data.extend(results)
                            data_urls.extend([result.get("meta_data").get("url") for result in results])
                    except Exception as e:
                        logger.error(f"Failed to process youtube video {video}: {e}")

        _add_youtube_channel()
        doc_id = hashlib.sha256((youtube_url + ", ".join(data_urls)).encode()).hexdigest()
        return {
            "doc_id": doc_id,
            "data": data,
        }



import hashlib
import json
import os
import re
from typing import Union

import requests

from mem0x.bootstrap.loaders.base_loader import BaseLoader
from mem0x.bootstrap.utils.misc import clean_string, is_valid_json_string


class JSONReader:
    def __init__(self) -> None:
        """Initialize the JSONReader."""
        pass

    @staticmethod
    def load_data(json_data: Union[dict, str]) -> list[str]:
        """Load data from a JSON structure.

        Args:
            json_data (Union[dict, str]): The JSON data to load.

        Returns:
            list[str]: A list of strings representing the leaf nodes of the JSON.
        """
        if isinstance(json_data, str):
            json_data = json.loads(json_data)
        else:
            json_data = json_data

        json_output = json.dumps(json_data, indent=0)
        lines = json_output.split("\n")
        useful_lines = [line for line in lines if not re.match(r"^[{}\[\],]*$", line)]
        return ["\n".join(useful_lines)]


VALID_URL_PATTERN = (
    "^https?://(?:www\.)?(?:\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}|[a-zA-Z0-9.-]+)(?::\d+)?/(?:[^/\s]+/)*[^/\s]+\.json$"
)


class JSONLoader(BaseLoader):
    @staticmethod
    def _check_content(content):
        if not isinstance(content, str):
            raise ValueError(
                "Invaid content input. \
                If you want to upload (list, dict, etc.), do \
                    `json.dump(data, indent=0)` and add the stringified JSON."
            )

    @staticmethod
    def load_data(content):
        """Load a json file. Each data point is a key value pair."""

        JSONLoader._check_content(content)
        loader = JSONReader()

        data = []
        data_content = []

        content_url_str = content

        if os.path.isfile(content):
            with open(content, "r", encoding="utf-8") as json_file:
                json_data = json.load(json_file)
        elif re.match(VALID_URL_PATTERN, content):
            response = requests.get(content)
            if response.status_code == 200:
                json_data = response.json()
            else:
                raise ValueError(
                    f"Loading data from the given url: {content} failed. \
                    Make sure the url is working."
                )
        elif is_valid_json_string(content):
            json_data = content
            content_url_str = hashlib.sha256((content).encode("utf-8")).hexdigest()
        else:
            raise ValueError(f"Invalid content to load json data from: {content}")

        docs = loader.load_data(json_data)
        for doc in docs:
            text = doc if isinstance(doc, str) else doc["text"]
            doc_content = clean_string(text)
            data.append({"content": doc_content, "meta_data": {"url": content_url_str}})
            data_content.append(doc_content)

        doc_id = hashlib.sha256((content_url_str + ", ".join(data_content)).encode()).hexdigest()
        return {"doc_id": doc_id, "data": data}



import hashlib
import logging
from pathlib import Path
from typing import Any, Optional

from mem0x.bootstrap.config import AddConfig
from mem0x.bootstrap.data_formatter.data_formatter import DataFormatter
from mem0x.bootstrap.helpers.json_serializable import register_deserializable
from mem0x.bootstrap.loaders.base_loader import BaseLoader
from mem0x.bootstrap.loaders.text_file import TextFileLoader
from mem0x.bootstrap.utils.misc import detect_datatype

logger = logging.getLogger(__name__)


@register_deserializable
class DirectoryLoader(BaseLoader):
    """Load data from a directory."""

    def __init__(self, config: Optional[dict[str, Any]] = None):
        super().__init__()
        config = config or {}
        self.recursive = config.get("recursive", True)
        self.extensions = config.get("extensions", None)
        self.errors = []

    def load_data(self, path: str):
        directory_path = Path(path)
        if not directory_path.is_dir():
            raise ValueError(f"Invalid path: {path}")

        logger.info(f"Loading data from directory: {path}")
        data_list = self._process_directory(directory_path)
        doc_id = hashlib.sha256((str(data_list) + str(directory_path)).encode()).hexdigest()

        for error in self.errors:
            logger.warning(error)

        return {"doc_id": doc_id, "data": data_list}

    def _process_directory(self, directory_path: Path):
        data_list = []
        for file_path in directory_path.rglob("*") if self.recursive else directory_path.glob("*"):
            # don't include dotfiles
            if file_path.name.startswith("."):
                continue
            if file_path.is_file() and (not self.extensions or any(file_path.suffix == ext for ext in self.extensions)):
                loader = self._predict_loader(file_path)
                data_list.extend(loader.load_data(str(file_path))["data"])
            elif file_path.is_dir():
                logger.info(f"Loading data from directory: {file_path}")
        return data_list

    def _predict_loader(self, file_path: Path) -> BaseLoader:
        try:
            data_type = detect_datatype(str(file_path))
            config = AddConfig()
            return DataFormatter(data_type=data_type, config=config)._get_loader(
                data_type=data_type, config=config.loader, loader=None
            )
        except Exception as e:
            self.errors.append(f"Error processing {file_path}: {e}")
            return TextFileLoader()



import hashlib
import logging
import time
from typing import Any, Optional

import requests

from mem0x.bootstrap.loaders.base_loader import BaseLoader
from mem0x.bootstrap.utils.misc import clean_string

logger = logging.getLogger(__name__)


class DiscourseLoader(BaseLoader):
    def __init__(self, config: Optional[dict[str, Any]] = None):
        super().__init__()
        if not config:
            raise ValueError(
                "DiscourseLoader requires a config."  # noqa: E501
            )

        self.domain = config.get("domain")
        if not self.domain:
            raise ValueError(
                "DiscourseLoader requires a domain."  # noqa: E501
            )

    def _check_query(self, query):
        if not query or not isinstance(query, str):
            raise ValueError(
                "DiscourseLoader requires a query."  # noqa: E501
            )

    def _load_post(self, post_id):
        post_url = f"{self.domain}posts/{post_id}.json"
        response = requests.get(post_url)
        try:
            response.raise_for_status()
        except Exception as e:
            logger.error(f"Failed to load post {post_id}: {e}")
            return
        response_data = response.json()
        post_contents = clean_string(response_data.get("raw"))
        metadata = {
            "url": post_url,
            "created_at": response_data.get("created_at", ""),
            "username": response_data.get("username", ""),
            "topic_slug": response_data.get("topic_slug", ""),
            "score": response_data.get("score", ""),
        }
        data = {
            "content": post_contents,
            "meta_data": metadata,
        }
        return data

    def load_data(self, query):
        self._check_query(query)
        data = []
        data_contents = []
        logger.info(f"Searching data on discourse url: {self.domain}, for query: {query}")
        search_url = f"{self.domain}search.json?q={query}"
        response = requests.get(search_url)
        try:
            response.raise_for_status()
        except Exception as e:
            raise ValueError(f"Failed to search query {query}: {e}")
        response_data = response.json()
        post_ids = response_data.get("grouped_search_result").get("post_ids")
        for id in post_ids:
            post_data = self._load_post(id)
            if post_data:
                data.append(post_data)
                data_contents.append(post_data.get("content"))
            # Sleep for 0.4 sec, to avoid rate limiting. Check `https://meta.discourse.org/t/api-rate-limits/208405/6`
            time.sleep(0.4)
        doc_id = hashlib.sha256((query + ", ".join(data_contents)).encode()).hexdigest()
        response_data = {"doc_id": doc_id, "data": data}
        return response_data



import csv
import hashlib
from io import StringIO
from urllib.parse import urlparse

import requests

from mem0x.bootstrap.loaders.base_loader import BaseLoader


class CsvLoader(BaseLoader):
    @staticmethod
    def _detect_delimiter(first_line):
        delimiters = [",", "\t", ";", "|"]
        counts = {delimiter: first_line.count(delimiter) for delimiter in delimiters}
        return max(counts, key=counts.get)

    @staticmethod
    def _get_file_content(content):
        url = urlparse(content)
        if all([url.scheme, url.netloc]) and url.scheme not in ["file", "http", "https"]:
            raise ValueError("Not a valid URL.")

        if url.scheme in ["http", "https"]:
            response = requests.get(content)
            response.raise_for_status()
            return StringIO(response.text)
        elif url.scheme == "file":
            path = url.path
            return open(path, newline="", encoding="utf-8")  # Open the file using the path from the URI
        else:
            return open(content, newline="", encoding="utf-8")  # Treat content as a regular file path

    @staticmethod
    def load_data(content):
        """Load a csv file with headers. Each line is a document"""
        result = []
        lines = []
        with CsvLoader._get_file_content(content) as file:
            first_line = file.readline()
            delimiter = CsvLoader._detect_delimiter(first_line)
            file.seek(0)  # Reset the file pointer to the start
            reader = csv.DictReader(file, delimiter=delimiter)
            for i, row in enumerate(reader):
                line = ", ".join([f"{field}: {value}" for field, value in row.items()])
                lines.append(line)
                result.append({"content": line, "meta_data": {"url": content, "row": i + 1}})
        doc_id = hashlib.sha256((content + " ".join(lines)).encode()).hexdigest()
        return {"doc_id": doc_id, "data": result}



import os

from mem0x import Memory, ShortTermMemory
from mem0x.configs.base import MemoryConfig, ShortTermMemoryConfig


def read_file_content(file_path: str):
    with open(file_path, "r", encoding="utf-8") as file:
        content = file.read()
        return content
    return None


class MemoryInstance:
    def __init__(self):
        self.m_config = MemoryConfig()
        self.m = Memory(self.m_config)

        self.sm_config = ShortTermMemoryConfig()
        self.sm = ShortTermMemory(self.sm_config)

        custom_prompt_file_path = os.getenv("LONGTERM_CUSTOM_PROMPT_FILE_PATH")
        if custom_prompt_file_path is not None:
            self.longterm_custom_prompt = read_file_content(custom_prompt_file_path)
        else:
            self.longterm_custom_prompt = None

        custom_prompt_file_path = os.getenv("BOOTSTRAP_CUSTOM_PROMPT_FILE_PATH")
        if custom_prompt_file_path is not None:
            self.bootstrap_custom_prompt = read_file_content(custom_prompt_file_path)
        else:
            self.bootstrap_custom_prompt = None


memoryInstance = MemoryInstance()



import hashlib

from mem0x.bootstrap.helpers.json_serializable import register_deserializable
from mem0x.bootstrap.loaders.base_loader import BaseLoader


@register_deserializable
class RSSFeedLoader(BaseLoader):
    """Loader for RSS Feed."""

    def load_data(self, url):
        """Load data from a rss feed."""
        output = self.get_rss_content(url)
        doc_id = hashlib.sha256((str(output) + url).encode()).hexdigest()
        return {
            "doc_id": doc_id,
            "data": output,
        }

    @staticmethod
    def serialize_metadata(metadata):
        for key, value in metadata.items():
            if not isinstance(value, (str, int, float, bool)):
                metadata[key] = str(value)

        return metadata

    @staticmethod
    def get_rss_content(url: str):
        try:
            from langchain_community.document_loaders import \
                RSSFeedLoader as LangchainRSSFeedLoader
        except ImportError:
            raise ImportError(
                """RSSFeedLoader file requires extra dependencies.
                Install with `pip install feedparser==6.0.10 newspaper3k==0.2.8 listparser==0.19`"""
            ) from None

        output = []
        loader = LangchainRSSFeedLoader(urls=[url])
        data = loader.load()

        for entry in data:
            metadata = RSSFeedLoader.serialize_metadata(entry.metadata)
            metadata.update({"url": url})
            output.append(
                {
                    "content": entry.page_content,
                    "meta_data": metadata,
                }
            )

        return output



import sqlite3
import uuid
from datetime import datetime


class SQLiteManager:
    def __init__(self, db_path=":memory:"):
        self.connection = sqlite3.connect(db_path, check_same_thread=False)
        self._create_history_table()

    def _create_history_table(self):
        with self.connection:
            self.connection.execute(
                """
                CREATE TABLE IF NOT EXISTS history (
                    id TEXT PRIMARY KEY,
                    memory_id TEXT,
                    old_memory TEXT,
                    new_memory TEXT,
                    event TEXT,
                    created_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
                    updated_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
                    is_deleted INTEGER
                )
            """
            )
            self.connection.execute(
                """
                CREATE INDEX IF NOT EXISTS idx_history ON history("id")
            """
            )

    def add_history(
        self,
        memory_id,
        old_memory,
        new_memory,
        event,
        created_at=None,
        updated_at=None,
        is_deleted=0,
    ):
        with self.connection:
            self.connection.execute(
                """
                INSERT INTO history (id, memory_id, old_memory, new_memory, event, created_at, updated_at, is_deleted)
                VALUES (?, ?, ?, ?, ?, ?, ?, ?)
            """,
                (
                    str(uuid.uuid4()),
                    memory_id,
                    old_memory,
                    new_memory,
                    event,
                    created_at or datetime.now(),
                    updated_at or datetime.now(),
                    is_deleted,
                ),
            )

    def get_history(self, memory_id):
        cursor = self.connection.execute(
            """
            SELECT id, memory_id, old_memory, new_memory, event, created_at, updated_at
            FROM history
            WHERE memory_id = ?
            ORDER BY updated_at ASC
        """,
            (memory_id,),
        )
        rows = cursor.fetchall()
        return [
            {
                "id": row[0],
                "memory_id": row[1],
                "old_memory": row[2],
                "new_memory": row[3],
                "event": row[4],
                "created_at": row[5],
                "updated_at": row[6],
            }
            for row in rows
        ]

    def reset(self):
        with self.connection:
            self.connection.execute("DROP INDEX IF EXISTS idx_history")
            self.connection.execute("DROP TABLE IF EXISTS history")
            self._create_history_table()



import importlib

from mem0x.configs.embeddings.base import BaseEmbedderConfig
from mem0x.configs.llms.base import BaseLlmConfig


def load_class(class_type):
    module_path, class_name = class_type.rsplit(".", 1)
    module = importlib.import_module(module_path)
    return getattr(module, class_name)


class LlmFactory:
    provider_to_class = {
        "gemini": "mem0x.llms.gemini.GeminiLLM",
    }

    @classmethod
    def create(cls, provider_name, config):
        class_type = cls.provider_to_class.get(provider_name)
        if class_type:
            llm_instance = load_class(class_type)
            base_config = BaseLlmConfig(**config)
            return llm_instance(base_config)
        else:
            raise ValueError(f"Unsupported Llm provider: {provider_name}")


class EmbedderFactory:
    provider_to_class = {
        "gemini": "mem0x.embeddings.gemini.GoogleGenAIEmbedding",
        "vertexai": "mem0x.embeddings.vertexai.VertexAIEmbedding",
    }

    @classmethod
    def create(cls, provider_name, config):
        class_type = cls.provider_to_class.get(provider_name)
        if class_type:
            embedder_instance = load_class(class_type)
            base_config = BaseEmbedderConfig(**config)
            return embedder_instance(base_config)
        else:
            raise ValueError(f"Unsupported Embedder provider: {provider_name}")


class VectorStoreFactory:
    provider_to_class = {
        "pgvector": "mem0x.vector_stores.pgvector.PGVector",
        "redis": "mem0x.vector_stores.redis.RedisDB",
    }

    @classmethod
    def create(cls, provider_name, config):
        class_type = cls.provider_to_class.get(provider_name)
        if class_type:
            if not isinstance(config, dict):
                config = config.model_dump()
            vector_store_instance = load_class(class_type)
            return vector_store_instance(**config)
        else:
            raise ValueError(f"Unsupported VectorStore provider: {provider_name}")
